# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
from collections import defaultdict
from typing import Dict
import numpy as np
from msprobe.core.common.log import logger
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import FileOpen, load_yaml
from msprobe.core.common.framework_adapter import FmkAdp

# both weights and bias are partitioned in column parallel
COLUMN_PARALLEL_PARAMS = ['linear_qkv', 'linear_fc1', 'word_embeddings.weight', 'output_layer.weight'] 
# only weights are partitioned in column parallel
ROW_PARALLEL_PARAMS = ['linear_fc2.weight', 'linear_proj.weight']
ARGS = 'args'
LAYER_IDX_PATTERN = re.compile('layers\.(\d+)\.')
EXPERT_IDX_PATTERN = re.compile('experts\.(\d+)\.')
ITER_DIR_PATTERN = re.compile('iter_([\d]{7})')


@recursion_depth_decorator('')
def _get_parameter(weights, prefix=''):
    for k, v in weights.items():
        name = Const.SEP.join([prefix, k]).strip(Const.SEP)
        if isinstance(v, dict):
            yield from _get_parameter(v, prefix=name)
        elif FmkAdp.is_tensor(v):
            yield name, FmkAdp.asnumpy(v)


def _map_to_mcore_local_names(param_name: str) -> str:
    """Map parameter names to mcore + local transformer implementation names."""
    mcore_local_map = load_yaml(os.path.join(os.path.dirname(__file__), 'name_mapping.yaml'))
    for other_name, mcore_local_name in mcore_local_map.items():
        param_name = param_name.replace(other_name, mcore_local_name)
    
    return param_name


def _parse_real_layer_idx(param_name, num_layers_per_stage, pp_size, pp_rank):
    """Map local (virtual) pipeline stage layer index to global layer index.
    
    For virtual pipeline parallel, each pipeline stage is further divided into virtual stages.
    The global layer index needs to account for both pipeline stage and virtual stage.
    
    Args:
        param_name (str): Parameter name containing layer index: layers.x.<submodule_name>/<vpp_stage>
        num_layers_per_stage (int): Number of layers per pipeline stage
        pp_size (int): Pipeline parallel size
        
    Returns:
        int: Global layer index accounting for both pipeline and virtual pipeline stages
    """
    # Extract local layer index from parameter name
    layer_match = re.search(LAYER_IDX_PATTERN, param_name)
    param_name, vpp_stage = param_name.split(Const.SCOPE_SEPARATOR)
    if not layer_match:
        return param_name
    
    local_layer_idx = int(layer_match.group(1))
    vpp_stage = int(vpp_stage)
    
    # Calculate global layer index based on pipeline stage and virtual stage
    real_layer_idx = local_layer_idx + (pp_size * vpp_stage + pp_rank) * num_layers_per_stage
    
    return param_name.replace(f'layers.{local_layer_idx}', f'layers.{real_layer_idx}')


def _parse_real_expert_idx(param_name, num_experts_per_rank, exp_rank):
    """Map local expert index to global expert index.
    
    For expert parallel, experts are distributed across ranks. This function maps
    the local expert index on a rank to its global index across all ranks.
    
    Args:
        param_name (str): Parameter name containing local expert index
        num_experts_per_rank (int): Number of experts on each rank
        exp_rank (int): Expert parallel rank
        
    Returns:
        str: Parameter name with local expert index replaced by global expert index
    """
    # Extract local layer index from parameter name
    expert_match = re.search(EXPERT_IDX_PATTERN, param_name)
    if not expert_match:
        return param_name
    
    local_expert_idx = int(expert_match.group(1))
    # Calculate global layer index based on pipeline stage and virtual stage
    real_experts_idx = local_expert_idx + exp_rank * num_experts_per_rank
    
    return param_name.replace(f'experts.{local_expert_idx}', f'experts.{real_experts_idx}')


def _consolidate_tp_weights(weights: Dict) -> Dict:
    """Consolidate weights from different tensor parallel ranks into combined tensors.
    
    Args:
        weights: Dictionary of weights with rank information in keys
        
    Returns:
        Dict: Consolidated weights without rank information
    """
    consolidated = {}
    for key, tensors in weights.items():    
        if any([name in key for name in COLUMN_PARALLEL_PARAMS]):
            # Column parallel - concatenate along input dimension (dim 0)
            combined = np.concatenate(tensors, axis=0)
        elif any([name in key for name in ROW_PARALLEL_PARAMS]):
            # Row parallel - concatenate along output dimension (dim 1)
            combined = np.concatenate(tensors, axis=1)
        else:
            # For other params, verify identical and use first
            if not all(np.allclose(tensors[0], t) for t in tensors[1:]):
                logger.warning(f"Inconsistent values for {key} across TP ranks")
            combined = tensors[0]
    
        consolidated[key] = combined
    return consolidated


def _parse_num_layers_per_stage(tp_partition):
    match = [re.findall(LAYER_IDX_PATTERN, key) for key in tp_partition.keys()]
    layer_idx = [int(i[0]) for i in match if i]
    if not layer_idx:
        return 1
    num_layers_per_pipeline_stage = max(layer_idx) + 1

    return num_layers_per_pipeline_stage


def parse_parallel_size(checkpoint_dir: str):
    """Parse tensor, pipeline and expert parallel sizes from checkpoint filenames.
    
    Args:
        checkpoint_dir (str): Directory containing checkpoint files
        
    Returns:
        Namespace
    """
    # Find all rank directories
    rank_dirs = [d for d in os.listdir(checkpoint_dir) if d.startswith('mp_rank_')]
    
    if not rank_dirs:
        raise ValueError(f"No checkpoint rank directories found in {checkpoint_dir}")
    
    ckpt = FmkAdp.load_checkpoint(
        os.path.join(checkpoint_dir, rank_dirs[0], 'model_optim_rng.pt'),
        to_cpu=True,
        weights_only=False)
    args = ckpt[ARGS]
    return (
        args.tensor_model_parallel_size, 
        args.pipeline_model_parallel_size, 
        args.expert_model_parallel_size, 
        args.num_experts
    )


def parse_iteration(checkpoint_path: str) -> Dict:
    """
    Parse the checkpoint iteration directory from a given checkpoint path.

    If the path is a top-level checkpoint directory, this function reads the
    'latest_checkpointed_iteration.txt' file to determine the latest iteration.
    If the path is already an iteration directory (e.g., 'iter_0000005'), it extracts
    the iteration number from the path.

    Args:
        checkpoint_path (str): Path to the checkpoint directory or iteration directory.

    Returns:
        str: The full path to the checkpoint directory for the determined iteration.

    Raises:
        ValueError: If the checkpoint directory for the determined iteration does not exist.
    """
    iteration = None
    tracker_file = os.path.join(checkpoint_path, "latest_checkpointed_iteration.txt")
    if os.path.exists(tracker_file):
        with FileOpen(tracker_file, 'r') as f:
            latest_iteration = f.read().strip()
            if latest_iteration != 'release':
                try:
                    iteration = int(latest_iteration)
                except Exception:
                    logger.warning(
                        f"The latest_checkpointed_iteration is supposed to be `release` or an int. \
                        But {latest_iteration} is found."
                    )
            checkpoint_path = os.path.join(checkpoint_path, f'iter_{iteration:07d}')
    else:
        match = re.findall(ITER_DIR_PATTERN, checkpoint_path)
        if match:
            iteration = int(match[0])

    # Checkpoint directory for this iteration
    logger.info(f"Loaded checkpoint from iteration {iteration}")
    
    if not os.path.exists(checkpoint_path):
        raise ValueError(f"Checkpoint directory not found: {checkpoint_path}")
    
    return checkpoint_path


def get_weights_from_state_dict(state_dict):
    weights = {}
    vpp_stage = 0
    if 'model' in state_dict:
        model_weights = state_dict['model']

        for key, value in _get_parameter(model_weights):
            key = _map_to_mcore_local_names(key)
            weights[f"{key}{Const.SCOPE_SEPARATOR}{vpp_stage}"] = value

    elif 'model0' in state_dict:
        #vpp enabled
        while f'model{vpp_stage}' in state_dict:
            model_weights = state_dict[f'model{vpp_stage}']
            for key, value in _get_parameter(model_weights):
                key = _map_to_mcore_local_names(key)
                weights[f"{key}{Const.SCOPE_SEPARATOR}{vpp_stage}"] = value
            vpp_stage += 1
    return weights


def load_megatron_weights(checkpoint_path: str) -> Dict:
    """Load Megatron parallel checkpoint weights into a single dictionary.
    
    Args:
        checkpoint_path (str): Base checkpoint directory path

    Returns:
        combined_weights: Dict with weights from all ranks, keys include rank info
    """
    try:
        import megatron
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError("No module named 'megatron', which is required to load a megatron ckpt") from e

    # Find latest iteration if not specified
    checkpoint_path = parse_iteration(checkpoint_path)

    # Parse parallel sizes from checkpoint directory structure
    tp_size, pp_size, exp_size, num_experts = parse_parallel_size(checkpoint_path)
    combined_weights = {}

    # Load checkpoints from all ranks
    for exp_rank in range(exp_size):
        num_layers_per_pipeline_stage = 0
        for pp_rank in range(pp_size):
            tp_partition = defaultdict(list)
            for tp_rank in range(tp_size):
                # Construct checkpoint path based on parallel ranks
                if pp_size > 1:
                    rank_dir = f'mp_rank_{tp_rank:02d}_{pp_rank:03d}'
                else:
                    rank_dir = f'mp_rank_{tp_rank:02d}'
                
                if exp_size > 1:
                    rank_dir = f'{rank_dir}_{exp_rank:03d}'
                
                ckpt_file = os.path.join(checkpoint_path, rank_dir, 'model_optim_rng.pt')
                try:
                    state_dict = FmkAdp.load_checkpoint(ckpt_file, to_cpu=True, weights_only=False)
                    partition = get_weights_from_state_dict(state_dict)
                    for key, weight in partition.items():
                        tp_partition[key].append(weight)
                    
                except Exception as load_error:
                    logger.warning(f"Error loading {ckpt_file}: {load_error}")
            
            if not tp_partition:
                raise ValueError('No state loaded.')
            
            if not num_layers_per_pipeline_stage:
                num_layers_per_pipeline_stage = _parse_num_layers_per_stage(tp_partition)

            consolidated_weight = _consolidate_tp_weights(tp_partition)
            for key, value in consolidated_weight.items(): 
                key = _parse_real_layer_idx(key, num_layers_per_pipeline_stage, pp_size, pp_rank)
                if num_experts:
                    key = _parse_real_expert_idx(key, num_experts // exp_size, exp_rank)
                combined_weights[key] = value

    logger.info(f"Found {len(combined_weights)} total parameters across all ranks")

    return combined_weights



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class is used to generate GUP dump data of the ONNX model.
"""
import os
import re
import time

import onnx
import onnxruntime
import numpy as np

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import check_file_or_directory_path, create_directory
from msprobe.infer.offline.compare.msquickcmp.common.dump_data import DumpData
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException
from msprobe.infer.offline.compare.msquickcmp.common.utils import InputShapeError, load_npy_from_buffer
from msprobe.infer.offline.compare.msquickcmp.adapter_cli.args_adapter import CmpArgsAdapter
from msprobe.infer.offline.compare.msquickcmp.common.convert import convert_bin_file_to_npy
from msprobe.infer.utils.file_open_check import ms_open
from msprobe.infer.utils.util import load_file_to_read_common_check
from msprobe.infer.utils.constants import TENSOR_MAX_SIZE


NODE_TYPE_TO_DTYPE_MAP = {
    "tensor(int)": np.int32,
    "tensor(int8)": np.int8,
    "tensor(int16)": np.int16,
    "tensor(int32)": np.int32,
    "tensor(int64)": np.int64,
    "tensor(uint8)": np.uint8,
    "tensor(uint16)": np.uint16,
    "tensor(uint32)": np.uint32,
    "tensor(uint64)": np.uint64,
    "tensor(float)": np.float32,
    "tensor(float16)": np.float16,
    "tensor(double)": np.double,
    "tensor(bool)": np.bool_,
    "tensor(complex64)": np.complex64,
    "tensor(complex128)": np.complex_,
}
MAX_PROTOBUF = 2000000000
ONNX_DTYPE = {1: np.float32, 2: np.float64}
MAX_FILE_NAME_LEN = 255


class OnnxDumpData(DumpData):
    """
    This class is used to generate dump data of the ONNX model.
    """

    def __init__(self, arguments: CmpArgsAdapter, npu_dump_npy_path=None):
        super().__init__()
        self.model_path = arguments.golden_path
        self.output_path = arguments.output_path
        self.input_data = arguments.input_data
        self.input_shape = arguments.input_shape
        self.dym_shape_range = arguments.dym_shape_range
        self.onnx_fusion_switch = arguments.onnx_fusion_switch
        self.cann_path = arguments.cann_path

        self._check_path_exists(self.model_path, extentions="onnx")

        self.input_shapes = utils.parse_input_shape(self.input_shape)
        self.data_dir, self.onnx_dump_data_dir, self.model_dir = self._create_dir()

        self.net_output, self.inputs_map, self.extend_inputs_map = {}, {}, {}
        self.origin_model, origin_model_contents = self._load_onnx(self.model_path)

        self.dump_model_with_inputs_path = self._new_model_save_path(self.model_path)
        self.model_with_inputs = self.origin_model
        self.model_with_inputs_session = self._load_session(self.model_path)

    @staticmethod
    def _check_input_shape_fix_value(op_name, model_shape, input_shape):
        message = "fixed input tensor dim not equal to model input dim." "tensor_name:%s, %s vs %s" % (
            op_name,
            str(input_shape),
            str(model_shape),
        )
        if len(model_shape) != len(input_shape):
            logger.error(message)
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_DATA_ERROR)
        for index, value in enumerate(model_shape):
            if value is None or isinstance(value, str):
                continue
            if input_shape[index] != value:
                logger.error(message)
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_DATA_ERROR)

    def generate_inputs_data(self, npu_dump_data_path=None, use_aipp=False):
        inputs_tensor_info = self._get_inputs_tensor_info()
        if use_aipp:
            if not npu_dump_data_path:
                logger.error("find no aipp op in dump data")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            self._check_path_exists(npu_dump_data_path)
            self.inputs_map = self._get_inputs_data_aipp(self.data_dir, inputs_tensor_info, npu_dump_data_path)
        else:
            self.inputs_map = self._get_inputs_data(inputs_tensor_info)

    def get_output_map(self, output_list):
        output_map, res_idx, output_list_size = {}, 0, len(output_list)
        for node in self.origin_model.graph.node:
            for node_output in node.output:
                output_map[node_output] = output_list[res_idx]
                res_idx += 1
                if res_idx >= output_list_size:
                    return output_map
        return output_map

    def get_input_map(self, input_map, output_list):
        output_map = self.get_output_map(output_list)
        input_map = self.augment_input_map(input_map, output_map)
        return input_map

    def augment_input_map(self, input_map, output_map):
        for temp in self.origin_model.graph.initializer:
            npy_data = load_npy_from_buffer(temp.raw_data, ONNX_DTYPE.get(temp.data_type), temp.dims)
            input_map[temp.name] = npy_data
        input_map = {**input_map, **output_map}
        return input_map
    
    def generate_dump_data(self, npu_dump_path=None, om_parser=None):
        self._modify_model_add_outputs_nodes(
            self.model_with_inputs, self.dump_model_with_inputs_path
        )
        session = self._load_session(self.dump_model_with_inputs_path)
        dump_bins = self._run_model(session, self.inputs_map)
        augment_inputs_map = self.get_input_map(self.inputs_map, dump_bins)
        net_output_node = [output_item.name for output_item in self.model_with_inputs_session.get_outputs()]
        self._save_dump_data(dump_bins, self.model_with_inputs, net_output_node, augment_inputs_map)

        return self.onnx_dump_data_dir

    def _load_session(self, model_path):
        options = onnxruntime.SessionOptions()
        if not self.onnx_fusion_switch:
            options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
        try:
            model_path = load_file_to_read_common_check(model_path)
            infersession = onnxruntime.InferenceSession(model_path, options)
        except Exception as e:
            logger.error(f"Please check onnx model can run in local env. Error: {e}")
            raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_MODEL_TYPE_ERROR)
        return infersession

    def _load_onnx(self, model_path):
        # model_path str -> read as bytes -> deserialize to onnx_model
        #                                 -> onnxruntime load as session
        model_path = load_file_to_read_common_check(model_path)
        with ms_open(model_path, "rb", max_size=TENSOR_MAX_SIZE) as ff:
            model_contents = ff.read()
        onnx_model = onnx.load_model(model_path)
        unique_index = 999
        name_set = set((node.name for node in onnx_model.graph.node))

        def get_unique_name(ori_name, index):
            new_name = f"{ori_name}_{index}"
            if new_name not in name_set:
                return new_name, index + 1
            else:
                return get_unique_name(ori_name, index + 1)

        for index, node in enumerate(onnx_model.graph.node):
            if node.name:
                continue
            new_name = node.op_type + "_" + str(index)
            if new_name in name_set:
                new_name, unique_index = get_unique_name(new_name, unique_index)
            name_set.add(new_name)
            node.name = new_name
        return onnx_model, model_contents

    def _new_model_save_path(self, origin_path):
        save_name = "new_" + os.path.basename(origin_path)
        return os.path.join(self.model_dir, save_name)

    def _create_dir(self):
        # create input directory
        data_dir = os.path.join(self.output_path, "input")
        create_directory(data_dir)

        # create dump_data/onnx directory
        onnx_dump_data_dir = os.path.join(self.output_path, "dump_data/onnx")
        create_directory(onnx_dump_data_dir)

        # create model directory
        model_dir = ""
        if self.dym_shape_range:
            model_relative_name = "../model"
        else:
            model_relative_name = "model"
            model_dir = os.path.join(self.output_path, model_relative_name)
            create_directory(model_dir)
        return data_dir, onnx_dump_data_dir, model_dir

    def _modify_model_add_outputs_nodes(self, onnx_model, save_path):
        del onnx_model.graph.output[:]

        onnx_model.graph.output.extend(
            onnx.ValueInfoProto(name=tensor_name)
            for node in onnx_model.graph.node
            for tensor_name in node.output
        )
        
        model_size = onnx_model.ByteSize()
        save_external_flag = model_size < 0 or model_size > MAX_PROTOBUF
        
        logger.debug(f"Modified model has size over 2G: {save_external_flag}")
        
        onnx.save_model(
            onnx_model, 
            save_path,
            save_as_external_data=save_external_flag
        )
        
        logger.info(f"Modified model has being saved successfully at: {os.path.abspath(save_path)}")
        
    def _get_inputs_tensor_info(self):
        inputs_tensor_info = []
        input_tensor_names = [item.name for item in self.model_with_inputs_session.get_inputs()]
        for _, tensor_name in enumerate(self.input_shapes):
            utils.check_input_name_in_model(input_tensor_names, tensor_name)
        for input_item in self.model_with_inputs_session.get_inputs():
            tensor_name = input_item.name
            tensor_type = input_item.type
            tensor_shape = tuple(input_item.shape)
            # skip extend inputs add by custom op
            if tensor_name in self.extend_inputs_map.keys():
                continue

            if utils.check_dynamic_shape(tensor_shape):
                if not self.input_shapes:
                    logger.error(
                        f"The dynamic shape {tensor_shape} are not supported. "
                        f"Please set '--input_shape' to fix the dynamic shape.")
                    raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            if self.input_shapes and tensor_name in self.input_shapes:
                input_shape = self.input_shapes.get(tensor_name)
                try:
                    number_shape = [int(dim) for dim in input_shape]
                except (ValueError, TypeError) as error:
                    logger.error(utils.get_shape_not_match_message(InputShapeError.FORMAT_NOT_MATCH, self.input_shape))
                    raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR) from error
                self._check_input_shape_fix_value(tensor_name, tensor_shape, number_shape)
                tensor_info = {"name": tensor_name, "shape": tuple(number_shape), "type": tensor_type}
                logger.info(f"Fix dynamic input shape of {tensor_name} to {number_shape}")
            else:
                tensor_info = {"name": tensor_name, "shape": tensor_shape, "type": tensor_type}
            inputs_tensor_info.append(tensor_info)
        logger.info(f"model inputs tensor info:\n{inputs_tensor_info}\n")
        return inputs_tensor_info

    def _get_inputs_data(self, inputs_tensor_info):
        names = [ii["name"] for ii in inputs_tensor_info]
        shapes = [ii["shape"] for ii in inputs_tensor_info]
        dtypes = [self._convert_to_numpy_type(ii["type"]) for ii in inputs_tensor_info]

        bin_file_path_array = []
        if "" == self.input_data:
            check_file_or_directory_path(os.path.realpath(self.data_dir), True)
            input_bin_files = os.listdir(self.data_dir)
            if len(input_bin_files) == 0:
                return self._generate_random_input_data(self.data_dir, names, shapes, dtypes)
            input_bin_files.sort(key=lambda file: int((re.findall("\\d+", file))[0]))
            bin_file_path_array = [os.path.join(self.data_dir, item) for item in input_bin_files]

        else:
            input_initial_path = self.input_data.split(",")
            for input_item in input_initial_path:
                input_item_path = os.path.realpath(input_item)
                if input_item_path.endswith('.bin'):
                    bin_file_path_array.append(input_item_path)
                else:
                    utils.get_input_path(input_item_path, bin_file_path_array)

        self._check_input_data_path(bin_file_path_array, inputs_tensor_info)
        return self._read_input_data(bin_file_path_array, names, shapes, dtypes)

    def _get_inputs_data_aipp(self, data_dir, inputs_tensor_info, npu_dump_data_path):
        inputs_map = {}
        aipp_input = []
        for bin_file in os.listdir(npu_dump_data_path):
            if bin_file.startswith("Aipp"):
                aipp_input.append(os.path.join(npu_dump_data_path, bin_file))
        if len(aipp_input) != len(inputs_tensor_info):
            logger.error("lengths of aipp_input and input_tensor_info unequal, please check.")
            raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INDEX_OUT_OF_BOUNDS_ERROR)
        for i, tensor_info in enumerate(inputs_tensor_info):
            convert_bin_file_to_npy(aipp_input[i], os.path.join(self.output_path, "input"), self.cann_path)
            aipp_output_path = (os.path.join(self.output_path, "input", aipp_input[i].rsplit("/", 1)[1]) +
                                ".output.0.npy")
            aipp_output_path = load_file_to_read_common_check(aipp_output_path)
            aipp_output = np.load(aipp_output_path)
            nchw_prod = np.prod(tensor_info["shape"])
            nchwc_prod_without_c1 = np.prod(aipp_output.shape[:-1])
            try:
                c0 = int(nchw_prod / nchwc_prod_without_c1)
            except ZeroDivisionError as e:
                logger.error(f"Aipp output has wrong shape, file path: {aipp_output_path}")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_DATA_ERROR) from e
            onnx_input = aipp_output[..., :c0].transpose((0, 4, 2, 3, 1)).squeeze(-1).astype(np.float32)
            inputs_map[tensor_info["name"]] = onnx_input
        return inputs_map

    def _convert_to_numpy_type(self, tensor_type):
        numpy_data_type = NODE_TYPE_TO_DTYPE_MAP.get(tensor_type)
        if numpy_data_type:
            return numpy_data_type
        else:
            logger.error(f"unsupported tensor type: {tensor_type}")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_TENSOR_TYPE_ERROR)

    def _run_model(self, session, inputs_map):
        outputs_name = [node.name for node in session.get_outputs()]
        return session.run(outputs_name, inputs_map)

    def _save_dump_data(self, dump_bins, old_onnx_model, net_output_node, input_map):
        res_idx = 0
        file_name_map = []
        for node in old_onnx_model.graph.node:
            #存储onnx的输入dump数据
            for i, node_input in enumerate(node.input):
                file_name = self._generate_dump_data_file_name("input_" + node.name, i)
                if len(file_name) > MAX_FILE_NAME_LEN:
                    new_file_name = str(round(time.time() * 1e6)) + str(len(file_name_map)) + ".npy"
                    file_name_map.append(f"{new_file_name},{file_name}\n")
                    file_name = new_file_name
                file_path = os.path.join(self.onnx_dump_data_dir, file_name)
                if input_map.get(node_input) is not None:
                    np.save(file_path, input_map.get(node_input))
            for j, output in enumerate(node.output):
                file_name = self._generate_dump_data_file_name(node.name, j)
                if len(file_name) > MAX_FILE_NAME_LEN:
                    new_file_name = str(round(time.time() * 1e6)) + str(len(file_name_map)) + ".npy"
                    file_name_map.append(f"{new_file_name},{file_name}\n")
                    file_name = new_file_name

                file_path = os.path.join(self.onnx_dump_data_dir, file_name)
                if output in net_output_node:
                    self.net_output[net_output_node.index(output)] = file_path
                if res_idx <= len(dump_bins) - 1:
                    np.save(file_path, dump_bins[res_idx])
                else:
                    logger.error("res_idx out of bounds of dump_bins and can not save, please check.")
                res_idx += 1

        if len(file_name_map) > 0:
            mapping_file_path = os.path.join(self.onnx_dump_data_dir, "mapping.csv")
            with ms_open(mapping_file_path, mode="w") as map_file:
                map_file.writelines(file_name_map)

        for key, value in self.net_output.items():
            logger.info(f"net_output node is:{key}, file path is {value}")
        logger.info("dump data success")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import multiprocessing
from multiprocessing.shared_memory import SharedMemory
import random
import time
import atexit
import os

from msprobe.core.common.log import logger


def is_main_process():
    return multiprocessing.current_process().name == 'MainProcess'


class GlobalLock:
    def __init__(self):
        self.name = self.get_lock_name()
        try:
            self._shm = SharedMemory(create=False, name=self.name)
            time.sleep(random.randint(0, 500) / 10000) # 等待随机时长以避免同时获得锁
        except FileNotFoundError:
            try:
                self._shm = SharedMemory(create=True, name=self.name, size=1)
                self._shm.buf[0] = 0
                logger.debug(f'{self.name} is created.')
            except FileExistsError:
                self.__init__()

    @classmethod
    def get_lock_name(cls):
        if is_main_process():
            return f'global_lock_{os.getpid()}'
        return f'global_lock_{os.getppid()}'

    @classmethod
    def is_lock_exist(cls):
        try:
            SharedMemory(create=False, name=cls.get_lock_name()).close()
            return True
        except FileNotFoundError:
            return False

    def cleanup(self):
        self._shm.close()
        if is_main_process():
            try:
                self._shm.unlink()
                logger.debug(f'{self.name} is unlinked.')
            except FileNotFoundError:
                logger.warning(f'{self.name} has already been unlinked.')

    def acquire(self, timeout=180):
        """
        acquire global lock, default timeout is 3 minutes.

        :param float timeout: timeout(seconds), default value is 180.
        """
        start = time.time()
        while time.time() - start < timeout:
            if self._shm.buf[0] == 0:
                self._shm.buf[0] = 1
                return
            time.sleep(random.randint(10, 500) / 10000)  # 自旋，等待1-50ms
        self._shm.buf[0] = 1

    def release(self):
        self._shm.buf[0] = 0


global_lock = GlobalLock()
atexit.register(global_lock.cleanup)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
from collections import namedtuple
import importlib

import torch

try:
    import torch_npu
except ImportError:
    IS_GPU = True
else:
    IS_GPU = False

from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import save_pt
from msprobe.core.common.file_utils import create_directory
from msprobe.core.common.const import Const
from msprobe.core.common.utils import CompareException

ApiData = namedtuple('ApiData', ['name', 'args', 'kwargs', 'result', 'step', 'rank'],
                     defaults=['unknown', None, None, None, 0, 0])


class DumpException(CompareException):
    pass


def check_object_type(check_object, allow_type):
    """
    Function Description:
        Check if the object belongs to a certain data type
    Parameter:
        check_object: the object to be checked
        allow_type: legal data type
    Exception Description:
        when invalid data throw exception
    """
    if not isinstance(check_object, allow_type):
        logger.error(f"{check_object} not of {allow_type} type")
        raise CompareException(CompareException.INVALID_DATA_ERROR)


class SoftlinkCheckException(Exception):
    pass


def check_need_convert(api_name):
    convert_type = None
    for key, value in Const.CONVERT_API.items():
        if api_name not in value:
            continue
        else:
            convert_type = key
    return convert_type


def cross_entropy_process(api_info_dict):
    """
    Function Description:
        Preprocesses the cross_entropy API information.
    Parameter:
        api_info_dict: argument of the API.
    Return api_info_dict:
        api_info_dict: Processed argument of the API.
    """
    if 'input_args' in api_info_dict and len(api_info_dict['input_args']) > 1 \
        and 'Min' in api_info_dict['input_args'][1]:
        if api_info_dict['input_args'][1]['Min'] <= 0:
            # The second argument in cross_entropy should be -100 or not less than 0
            api_info_dict['input_args'][1]['Min'] = 0
    return api_info_dict


def histc_process(api_info_dict):
    input_args = api_info_dict['input_args']
    if input_args and input_args[0].get('dtype'):
        dtype = input_args[0]['dtype']
        if dtype in Const.TORCH_INT_DTYPE:
            api_info_dict['input_args'][0]['dtype'] = Const.TORCH_FLOAT32
    return api_info_dict


API_PROCESS_MAP = {
    'cross_entropy': cross_entropy_process,
    'histc': histc_process
}


def api_info_preprocess(api_name, api_info_dict):
    """
    Function Description:
        Preprocesses the API information.
    Parameter:
        api_name: Name of the API.
        api_info_dict: argument of the API.
    Return api_info_dict:
        convert_type: Type of conversion.
        api_info_dict: Processed argument of the API.
    """
    convert_type = check_need_convert(api_name)
    if api_name in API_PROCESS_MAP:
        api_info_dict = API_PROCESS_MAP[api_name](api_info_dict)
    return convert_type, api_info_dict


def initialize_save_path(save_path, dir_name):
    data_path = os.path.join(save_path, dir_name)
    create_directory(data_path)
    return data_path


def get_full_data_path(data_path, real_data_path):
    if not data_path:
        return data_path
    full_data_path = os.path.join(real_data_path, data_path)
    return os.path.realpath(full_data_path)


class UtDataProcessor:
    def __init__(self, save_path):
        self.save_path = save_path
        self.index = 0

    def save_tensors_in_element(self, api_name, element):
        self.index = 0
        self._save_recursive(api_name, element)

    def _save_recursive(self, api_name, element, depth=0):
        if depth > Const.MAX_DEPTH:
            logger.error(f"Maximum depth of {Const.MAX_DEPTH} exceeded for {api_name}")
            raise DumpException(DumpException.RECURSION_LIMIT_ERROR)
        if isinstance(element, torch.Tensor):
            api_args = api_name + Const.SEP + str(self.index)
            create_directory(self.save_path)
            file_path = os.path.join(self.save_path, f'{api_args}.pt')
            try:
                tensor = element.contiguous().detach().cpu()
            except Exception as err:
                logger.error(f"Failed to transfer tensor to cpu for {api_args}")
                raise DumpException(DumpException.INVALID_DATA_ERROR) from err
            save_pt(tensor, file_path)
            self.index += 1
        elif element is None or isinstance(element, (bool, int, float, str, slice)):
            self.index += 1
        elif isinstance(element, (list, tuple)):
            for item in element:
                self._save_recursive(api_name, item, depth=depth + 1)
        elif isinstance(element, dict):
            for value in element.values():
                self._save_recursive(api_name, value, depth=depth + 1)
        else:
            self.index += 1


def extract_basic_api_segments(api_full_name):
    """
    Function Description:
        Extract the name of the API.
    Parameter:
        api_full_name: Full name of the API. Example: torch.matmul.0, torch.linalg.inv.0
    Return:
        api_type: Type of api. Example: torch, tensor, etc.
        api_name: Name of api. Example: matmul, linalg.inv, etc.
    """
    api_type = None
    api_parts = api_full_name.split(Const.SEP)
    api_parts_length = len(api_parts)
    if api_parts_length == Const.THREE_SEGMENT:
        api_type, api_name, _ = api_parts
    elif api_parts_length == Const.FOUR_SEGMENT:
        api_type, prefix, api_name, _ = api_parts
        api_name = Const.SEP.join([prefix, api_name])
    else:
        api_name = None
    return api_type, api_name


def extract_detailed_api_segments(full_api_name_with_direction_status):
    """
    Function Description:
        Extract the name of the API.
    Parameter:
        full_api_name_with_direction_status: Full name of the API. Example: torch.matmul.0.forward.output.0
    Return:
        api_name: Name of api. Example: matmul, mul, etc.
        full_api_name: Full name of api. Example: torch.matmul.0
        direction_status: Direction status of api. Example: forward, backward, etc.
    """
    api_type = None
    prefix = None
    api_name = None
    direction_status = None
    api_parts = full_api_name_with_direction_status.split(Const.SEP)
    api_parts_length = len(api_parts)
    if api_parts_length == Const.SIX_SEGMENT:
        api_type, api_name, api_order, direction_status, _, _ = api_parts
        full_api_name = Const.SEP.join([api_type, api_name, api_order])
    elif api_parts_length == Const.SEVEN_SEGMENT:
        api_type, prefix, api_name, api_order, direction_status, _, _ = api_parts
        full_api_name = Const.SEP.join([api_type, prefix, api_name, api_order])
        api_name = Const.SEP.join([prefix, api_name])
    else:
        full_api_name = None
    return api_name, full_api_name, direction_status


def get_module_and_atttribute_name(attribute):
    '''
    Function Description:
        Get the module and attribute name.
    Parameter:
        name: Attribute of a module. Example: torch.float16
    Return:
        module_name: Name of the module. Example: torch.
        attribute_name: Name of the attribute. Example: float16.
    '''
    try:
        module_name, attribute_name = attribute.split(Const.SEP)
    except ValueError as e:
        logger.error(f"Failed to get module and attribute name from {attribute}")
        raise CompareException(CompareException.INVALID_DATA_ERROR) from e
    return module_name, attribute_name


def get_attribute(module_name, attribute_name):
    '''
    Function Description:
        Get the attribute of the module.
    Parameter:
        module_name: Name of the module.
        attribute_name: Name of the attribute.
    '''
    attribute = None
    if module_name not in Const.MODULE_WHITE_LIST:
        logger.error(f"Module {module_name} is not in white list")
        raise CompareException(CompareException.INVALID_DATA_ERROR)
    try:
        module = importlib.import_module(module_name)
        attribute = getattr(module, attribute_name)
    except (ImportError, AttributeError) as e:
        logger.error(f"Failed to get attribute {attribute_name} from module {module_name}: {e}")
        raise CompareException(CompareException.INVALID_ATTRIBUTE_ERROR) from e
    return attribute


def is_dtype_fp8(dtype):
    """
    Function Description:
        Check if the data type is float8.
    Parameter:
        dtype: Data type (torch.dtype or string).
    Return:
        True or False.
    """
    # 处理字符串类型的 dtype
    if isinstance(dtype, str):
        return dtype in ["torch.float8_e4m3fn", "torch.float8_e5m2"]

    # 处理 torch.dtype 类型
    return dtype in [torch.float8_e4m3fn, torch.float8_e5m2]


def is_dtype_hif8(dtype):
    """
    Function Description:
        Check if the data type is HiFloat8Tensor.
    Parameter:
        dtype: Data type (string).
    Return:
        True or False.
    """
    # 处理字符串类型的 dtype
    if isinstance(dtype, str):
        dtype_str = dtype
    # 处理类对象或 dtype 对象
    else:
        dtype_str = str(dtype)

    # 检查是否匹配 HiFloat8Tensor 的字符串表示
    return (
            dtype_str == "<class 'torch_npu.utils.hif8_tensor.HiFloat8Tensor'>" or
            dtype_str == "torch_npu.HiFloat8Tensor"
    )


def is_dtype_fp8_or_hif8(dtype):
    """
    Function Description:
        Check if the data type is FP8 (native or HiFloat8 variant).
    Parameters:
        dtype: Data type (torch.dtype, class, or string).
    Returns:
        True if dtype is FP8 or HiFloat8, False otherwise.
    """
    return is_dtype_fp8(dtype) or is_dtype_hif8(dtype)


def is_hifloat8_tensor(tensor):
    if not IS_GPU and hasattr(torch_npu, "HiFloat8Tensor") and isinstance(tensor, torch_npu.HiFloat8Tensor):
        return True
    return False


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import time
import threading
from copy import deepcopy
from dataclasses import dataclass
from multiprocessing import cpu_count, Pool, Manager
from typing import Callable, Optional

from tqdm import tqdm
from msprobe.core.common.file_utils import (check_file_type, create_directory, FileChecker,
                                            check_file_or_directory_path, load_json)
from msprobe.core.common.const import FileCheckConst, Const
from msprobe.core.common.utils import CompareException, get_dump_mode
from msprobe.visualization.compare.graph_comparator import GraphComparator
from msprobe.visualization.utils import GraphConst, check_directory_content, SerializableArgs, load_parallel_param, \
    sort_rank_number_strings, validate_parallel_param, get_step_or_rank_int, \
    monitor_progress, ProgressInfo, calculate_list, get_log_msg_wrapper
from msprobe.visualization.builder.graph_builder import GraphBuilder, GraphExportConfig, GraphInfo, BuildGraphTaskInfo
from msprobe.core.common.log import logger, BaseLogger
from msprobe.visualization.graph.node_colors import NodeColors
from msprobe.core.compare.layer_mapping import generate_api_mapping_by_layer_mapping
from msprobe.core.compare.utils import check_and_return_dir_contents
from msprobe.core.common.utils import detect_framework_by_dump_json
from msprobe.visualization.graph.distributed_analyzer import distributed_analyse
from msprobe.visualization.builder.graph_merger import GraphMerger
from msprobe.visualization.db_utils import post_process_db

current_time = time.strftime("%Y%m%d%H%M%S")
build_output_db_name = f'build_{current_time}.vis.db'
compare_output_db_name = f'compare_{current_time}.vis.db'


def _compare_graph(graph_n: GraphInfo, graph_b: GraphInfo, input_param, args, pbar_info=None):
    dump_path_param = {
        'npu_path': graph_n.data_path,
        'bench_path': graph_b.data_path,
        'stack_path': graph_n.stack_path,
        'is_print_compare_log': input_param.get("is_print_compare_log", False)
    }
    mapping_dict = {}
    if args.layer_mapping:
        try:
            mapping_dict = generate_api_mapping_by_layer_mapping(graph_n.data_path, graph_b.data_path,
                                                                 args.layer_mapping)
        except Exception:
            logger.warning('The layer mapping file parsing failed, please check file format, mapping is not effective.')
    is_cross_framework = detect_framework_by_dump_json(graph_n.data_path) != \
                         detect_framework_by_dump_json(graph_b.data_path)
    if is_cross_framework and not args.layer_mapping:
        logger.error('The cross_frame graph comparison failed. '
                     'Please specify -lm or --layer_mapping when performing cross_frame graph comparison.')
        raise CompareException(CompareException.CROSS_FRAME_ERROR)

    graph_comparator = GraphComparator([graph_n.graph, graph_b.graph], dump_path_param, args, is_cross_framework,
                                       mapping_dict=mapping_dict, pbar_info=pbar_info)
    graph_comparator.compare()
    return graph_comparator


def _compare_graph_result(input_param, args, pbar_info=None):
    logger.info('Start building model graphs...')
    # 对两个数据进行构图
    graph_n = _build_graph_info(input_param.get('npu_path'), args, pbar_info=pbar_info)
    graph_b = _build_graph_info(input_param.get('bench_path'), args, pbar_info=pbar_info)
    logger.info('Model graphs built successfully, start comparing graphs...')
    # 基于graph、stack和data进行比较
    graph_comparator = _compare_graph(graph_n, graph_b, input_param, args, pbar_info=pbar_info)
    # 增加micro step标记
    micro_steps = graph_n.graph.paging_by_micro_step(graph_b.graph)
    # 开启溢出检测
    if args.overflow_check:
        graph_n.graph.overflow_check()
        graph_b.graph.overflow_check()

    return CompareGraphResult(graph_n.graph, graph_b.graph, graph_comparator, micro_steps)


def _export_compare_graph_result(args, result, pbar_info=None):
    graphs = [result.graph_n, result.graph_b]
    graph_comparator = result.graph_comparator
    micro_steps = result.micro_steps
    logger.info(f'Start exporting compare graph result, file name: {compare_output_db_name}...')
    output_db_path = os.path.join(args.output_path, compare_output_db_name)
    task = GraphConst.GRAPHCOMPARE_MODE_TO_DUMP_MODE_TO_MAPPING.get(graph_comparator.ma.compare_mode)
    export_config = GraphExportConfig(graphs[0], graphs[1], graph_comparator.ma.get_tool_tip(),
                                      NodeColors.get_node_colors(graph_comparator.ma.compare_mode), micro_steps, task,
                                      args.overflow_check, graph_comparator.ma.compare_mode, result.step, result.rank,
                                      args.step_list if hasattr(args, 'step_list') else [0],
                                      args.rank_list if hasattr(args, 'rank_list') else [0])
    try:
        GraphBuilder.to_db(output_db_path, export_config, pbar_info=pbar_info)
        logger.info(f'Exporting compare graph result successfully, the result file is saved in {output_db_path}')
        return ''
    except RuntimeError as e:
        logger.error(f'Failed to export compare graph result, file: {compare_output_db_name}, error: {e}')
        return compare_output_db_name


def _build_graph_info(dump_path, args, graph=None, pbar_info=None):
    construct_path = FileChecker(os.path.join(dump_path, GraphConst.CONSTRUCT_FILE), FileCheckConst.FILE,
                                 FileCheckConst.READ_ABLE).common_check()
    data_path = FileChecker(os.path.join(dump_path, GraphConst.DUMP_FILE), FileCheckConst.FILE,
                            FileCheckConst.READ_ABLE).common_check()
    stack_path = FileChecker(os.path.join(dump_path, GraphConst.STACK_FILE), FileCheckConst.FILE,
                             FileCheckConst.READ_ABLE).common_check()
    if not graph:
        graph = GraphBuilder.build(construct_path, data_path, stack_path, pbar_info=pbar_info)
    return GraphInfo(graph, construct_path, data_path, stack_path)


def _build_graph_result(dump_path, args, pbar_info=None):
    logger.info('Start building model graphs...')
    graph = _build_graph_info(dump_path, args, pbar_info=pbar_info).graph
    # 增加micro step标记
    micro_steps = graph.paging_by_micro_step()
    # 开启溢出检测
    if args.overflow_check:
        graph.overflow_check()
    return BuildGraphResult(graph, micro_steps)


def _run_build_graph_compare(input_param, args, nr, br, pbar_info=None):
    logger.info(f'Start building graph for {nr}...')
    graph_n = _build_graph_info(input_param.get('npu_path'), args, pbar_info=pbar_info)
    graph_b = _build_graph_info(input_param.get('bench_path'), args, pbar_info=pbar_info)
    logger.info(f'Building graph for {nr} finished.')
    return BuildGraphTaskInfo(graph_n, graph_b, nr, br, current_time)


def _run_build_graph_single(dump_ranks_path, rank, step, args, pbar_info=None):
    logger.info(f'Start building graph for {rank}...')
    dump_path = os.path.join(dump_ranks_path, rank)
    result = _build_graph_result(dump_path, args, pbar_info=pbar_info)
    if rank != Const.RANK:
        result.rank = get_step_or_rank_int(rank, True)
    logger.info(f'Building graph for step: {step}, rank: {rank} finished.')
    return result


def _run_build_graph_and_export(dump_ranks_path, rank, step, args, pbar_info=None):
    result = _run_build_graph_single(dump_ranks_path, rank, step, args, pbar_info)
    if step is not None:
        result.step = get_step_or_rank_int(step)
    create_directory(args.output_path)

    return _export_build_graph_result(args, result, pbar_info)


def _run_graph_compare(graph_task_info, input_param, args, pbar_info=None):
    logger.info(f'Start comparing data for {graph_task_info.npu_rank}...')
    graph_n = graph_task_info.graph_info_n
    graph_b = graph_task_info.graph_info_b
    nr = graph_task_info.npu_rank
    graph_comparator = _compare_graph(graph_n, graph_b, input_param, args, pbar_info=pbar_info)
    micro_steps = graph_n.graph.paging_by_micro_step(graph_b.graph)
    # 开启溢出检测
    if args.overflow_check:
        graph_n.graph.overflow_check()
        graph_b.graph.overflow_check()
    graph_result = CompareGraphResult(graph_n.graph, graph_b.graph, graph_comparator, micro_steps)
    if nr != Const.RANK:
        graph_result.rank = get_step_or_rank_int(nr, True)
    logger.info(f'Comparing data for {graph_task_info.npu_rank} finished.')
    return graph_result


def _export_build_graph_result(args, result, pbar_info=None):
    out_path = args.output_path
    graph = result.graph
    micro_steps = result.micro_steps
    overflow_check = args.overflow_check
    logger.info(f'Start exporting graph for {build_output_db_name}...')
    output_db_path = os.path.join(out_path, build_output_db_name)
    config = GraphExportConfig(graph, micro_steps=micro_steps, overflow_check=overflow_check, rank=result.rank,
                               step=result.step, rank_list=args.rank_list if hasattr(args, 'rank_list') else [0],
                               step_list=args.step_list if hasattr(args, 'step_list') else [0])
    try:
        GraphBuilder.to_db(output_db_path, config, pbar_info=pbar_info)
        logger.info(f'Model graph exported successfully, the result file is saved in {output_db_path}')
        return None
    except RuntimeError as e:
        logger.error(f'Failed to export model graph, file: {build_output_db_name}, error: {e}')
        return build_output_db_name


def is_real_data_compare(input_param, npu_ranks, bench_ranks):
    dump_rank_n = input_param.get('npu_path')
    dump_rank_b = input_param.get('bench_path')
    has_real_data = False
    for nr, br in zip(npu_ranks, bench_ranks):
        dump_path_param = {
            'npu_path': FileChecker(os.path.join(dump_rank_n, nr, GraphConst.DUMP_FILE), FileCheckConst.FILE,
                                    FileCheckConst.READ_ABLE).common_check(),
            'bench_path': FileChecker(os.path.join(dump_rank_b, br, GraphConst.DUMP_FILE), FileCheckConst.FILE,
                                      FileCheckConst.READ_ABLE).common_check()
        }
        has_real_data |= get_dump_mode(dump_path_param) == Const.ALL
    return has_real_data


def _mp_compare(input_param, serializable_args, nr, br, pbar_info=None):
    graph_task_info = _run_build_graph_compare(input_param, serializable_args, nr, br, pbar_info=pbar_info)
    return _run_graph_compare(graph_task_info, input_param, serializable_args, pbar_info=pbar_info)


def _mp_compare_and_export(input_param, args, rank, step, pbar_info=None):
    graph_result = _mp_compare(input_param, args, rank, rank, pbar_info=pbar_info)
    if step is not None:
        graph_result.step = get_step_or_rank_int(step)
    create_directory(args.output_path)

    return _export_compare_graph_result(args, graph_result, pbar_info=pbar_info)


def _compare_graph_ranks(input_param, args, step=None, pbar_info=None):
    dump_rank_n = input_param.get('npu_path')
    dump_rank_b = input_param.get('bench_path')
    npu_ranks = sort_rank_number_strings(check_and_return_dir_contents(dump_rank_n, Const.RANK))
    bench_ranks = sort_rank_number_strings(check_and_return_dir_contents(dump_rank_b, Const.RANK))
    if npu_ranks != bench_ranks:
        intersection_ranks = sort_rank_number_strings(list(set(npu_ranks) & set(bench_ranks)))
        if not intersection_ranks:
            logger.error('The ranks in the two runs are completely different. Unable to match the ranks.')
            raise CompareException(CompareException.INVALID_PATH_ERROR)
        npu_ranks = intersection_ranks
        bench_ranks = intersection_ranks
    args.rank_list = [get_step_or_rank_int(rank, True) for rank in npu_ranks]
    serializable_args = SerializableArgs(args)

    with Pool(processes=max(int((cpu_count() + 1) // 4), 1)) as pool:
        def err_call(err):
            logger.error(f'Error occurred while comparing graph ranks: {err}')

        if is_real_data_compare(input_param, npu_ranks, bench_ranks):
            # 真实数据模式，考虑到tensor比对过程会使用进程池启用多进程，为了避免嵌套进程池，graph比对使用串行
            compare_graph_results = []
            mp_task_dict = {}
            for nr, br in zip(npu_ranks, bench_ranks):
                input_param['npu_path'] = os.path.join(dump_rank_n, nr)
                input_param['bench_path'] = os.path.join(dump_rank_b, br)
                build_key = f'{step}_{nr}' if step else f'{nr}'
                input_param_copy = deepcopy(input_param)
                pbar_info_copy = PbarInfo.update_task_id(pbar_info, nr)
                mp_task_dict[build_key] = pool.apply_async(_run_build_graph_compare,
                                                           args=(input_param_copy, serializable_args, nr, br,
                                                                 pbar_info_copy),
                                                           error_callback=err_call)
            mp_res_dict = {k: v.get() for k, v in mp_task_dict.items()}
            for build_key, mp_res in mp_res_dict.items():
                if pbar_info:
                    if Const.REPLACEMENT_CHARACTER in build_key:
                        build_key = build_key.split(Const.REPLACEMENT_CHARACTER)[-1]
                    pbar_info.task_id = build_key
                compare_graph_results.append(_run_graph_compare(mp_res, input_param, serializable_args, pbar_info))
            if step is not None:
                for result in compare_graph_results:
                    result.step = get_step_or_rank_int(step)

            export_res_task_list = []
            create_directory(args.output_path)
            for result in compare_graph_results:
                export_res_task_list.append(pool.apply_async(_export_compare_graph_result,
                                                             args=(serializable_args, result),
                                                             error_callback=err_call))
            export_res_list = [res.get() for res in export_res_task_list]
        else:
            compare_graph_tasks = []
            for nr, br in zip(npu_ranks, bench_ranks):
                input_param['npu_path'] = os.path.join(dump_rank_n, nr)
                input_param['bench_path'] = os.path.join(dump_rank_b, br)
                input_param_copy = deepcopy(input_param)
                pbar_info_copy = PbarInfo.update_task_id(pbar_info, nr)
                compare_graph_tasks.append(pool.apply_async(_mp_compare_and_export,
                                                            args=(input_param_copy, serializable_args, nr, step,
                                                                  pbar_info_copy),
                                                            error_callback=err_call))
            export_res_list = [res.get() for res in compare_graph_tasks]
        if any(export_res_list):
            failed_names = list(filter(lambda x: x, export_res_list))
            logger.error(f'Unable to export compare graph results: {", ".join(failed_names)}.')
        else:
            logger.info('Successfully exported compare graph results.')


def _compare_graph_steps(input_param, args, pbar_info=None):
    dump_step_n = input_param.get('npu_path')
    dump_step_b = input_param.get('bench_path')

    npu_steps = calculate_list(dump_step_n, dump_step_b, Const.STEP)

    args.step_list = sorted([get_step_or_rank_int(step) for step in npu_steps])

    for i, folder_step in enumerate(npu_steps):
        logger.info(f'Start processing data for {folder_step}...')
        input_param['npu_path'] = os.path.join(dump_step_n, folder_step)
        input_param['bench_path'] = os.path.join(dump_step_b, folder_step)

        if pbar_info:
            pbar_info.step = i

        _compare_graph_ranks(input_param, args, step=folder_step, pbar_info=pbar_info) if not args.parallel_merge \
            else _compare_graph_ranks_parallel(input_param, args, step=folder_step, pbar_info=pbar_info)


def _build_graph_ranks_parallel(args, step=None, pbar_info=None):
    dump_ranks_path = os.path.join(args.target_path, step) if step is not None else args.target_path
    ranks = sort_rank_number_strings(check_and_return_dir_contents(dump_ranks_path, Const.RANK))
    serializable_args = SerializableArgs(args)
    with Pool(processes=max(int((cpu_count() + 1) // 4), 1)) as pool:
        def err_call(err):
            logger.error(f'Error occurred while comparing graph ranks: {err}')
            try:
                pool.close()
            except OSError as e:
                logger.error(f'Error occurred while terminating the pool: {e}')

        build_graph_tasks = []
        if pbar_info and pbar_info.step:
            PbarInfo.reset_progress_and_current_stage(pbar_info, ranks)
        for rank in ranks:
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, rank)
            build_graph_tasks.append(pool.apply_async(_run_build_graph_single,
                                                      args=(dump_ranks_path, rank, step, serializable_args,
                                                            pbar_info_copy), error_callback=err_call))
        build_graph_results = [task.get() for task in build_graph_tasks]

        if step is not None:
            for result in build_graph_results:
                result.step = get_step_or_rank_int(step)

        if args.parallel_params:
            validate_parallel_param(args.parallel_params[0], dump_ranks_path)
            build_graph_results = GraphMerger(build_graph_results, args.parallel_params[0],
                                              pbar_info=pbar_info).merge_graph()
            if pbar_info:
                PbarInfo.del_progress_dict_item(pbar_info, ranks,
                                                [f'{Const.RANK}{result.rank}' for result in build_graph_results])

        create_directory(args.output_path)
        export_build_graph_tasks = []
        serializable_args.rank_list = [result.rank for result in build_graph_results]
        for result in build_graph_results:
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, f'{Const.RANK}{result.rank}')
            export_build_graph_tasks.append(pool.apply_async(_export_build_graph_result,
                                                             args=(serializable_args, result, pbar_info_copy),
                                                             error_callback=err_call))
        export_build_graph_result = [task.get() for task in export_build_graph_tasks]
        if any(export_build_graph_result):
            failed_names = list(filter(lambda x: x, export_build_graph_result))
            logger.error(f'Unable to export build graph results: {failed_names}.')
        else:
            logger.info(f'Successfully exported build graph results.')


def _build_graph_ranks(args, step=None, pbar_info=None):
    dump_ranks_path = os.path.join(args.target_path, step) if step is not None else args.target_path
    ranks = sort_rank_number_strings(check_and_return_dir_contents(dump_ranks_path, Const.RANK))
    args.rank_list = [get_step_or_rank_int(rank, True) for rank in ranks]
    serializable_args = SerializableArgs(args)
    with Pool(processes=max(int((cpu_count() + 1) // 4), 1)) as pool:
        def err_call(err):
            logger.error(f'Error occurred while comparing graph ranks: {err}')

        tasks = []
        for rank in ranks:
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, rank)
            tasks.append(pool.apply_async(_run_build_graph_and_export,
                                          args=(dump_ranks_path, rank, step, serializable_args, pbar_info_copy),
                                          error_callback=err_call))
        results = [task.get() for task in tasks]
        if any(results):
            failed_names = list(filter(lambda x: x, results))
            logger.error(f'Unable to export build graph results: {failed_names}.')
        else:
            logger.info(f'Successfully exported build graph results.')


def _build_graph_steps(args, pbar_info=None):
    steps = sorted(check_and_return_dir_contents(args.target_path, Const.STEP))
    args.step_list = sorted([get_step_or_rank_int(step) for step in steps])

    for i, step in enumerate(steps):
        logger.info(f'Start processing data for {step}...')
        if pbar_info:
            pbar_info.step = i
        _build_graph_ranks(args, step, pbar_info=pbar_info) if not args.parallel_merge \
            else _build_graph_ranks_parallel(args, step, pbar_info=pbar_info)


def _compare_and_export_graph(graph_task_info, input_param, args, step=None, pbar_info=None):
    result = _run_graph_compare(graph_task_info, input_param, args, pbar_info=pbar_info)
    if step is not None:
        result.step = get_step_or_rank_int(step)
    return _export_compare_graph_result(args, result, pbar_info=pbar_info)


def _compare_graph_ranks_parallel(input_param, args, step=None, pbar_info=None):
    args.fuzzy_match = True
    npu_path = input_param.get('npu_path')
    bench_path = input_param.get('bench_path')
    ranks_n = sort_rank_number_strings(check_and_return_dir_contents(npu_path, Const.RANK))
    ranks_b = sort_rank_number_strings(check_and_return_dir_contents(bench_path, Const.RANK))
    parallel_params = args.parallel_params
    if len(parallel_params) != 2:
        raise RuntimeError('Parallel params error in compare graph!')
    validate_parallel_param(parallel_params[0], npu_path)
    validate_parallel_param(parallel_params[1], bench_path, '[Bench]')
    serializable_args = SerializableArgs(args)

    with Pool(processes=max(int((cpu_count() + 1) // 4), 1)) as pool:
        def err_call(err):
            logger.error(f'Error occurred while comparing graph ranks: {err}')
            try:
                pool.close()
            except OSError as e:
                logger.error(f'Error occurred while terminating the pool: {e}')

        # 1.并行构图
        build_graph_tasks_n = []
        build_graph_tasks_b = []
        if pbar_info and pbar_info.step:
            PbarInfo.reset_progress_and_current_stage(pbar_info, list(set(ranks_n) | set(ranks_b)))
        for rank in ranks_n:
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, rank)
            build_graph_tasks_n.append(pool.apply_async(_run_build_graph_single,
                                                        args=(npu_path, rank, step, serializable_args, pbar_info_copy),
                                                        error_callback=err_call))
        for rank in ranks_b:
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, rank)
            build_graph_tasks_b.append(pool.apply_async(_run_build_graph_single,
                                                        args=(bench_path, rank, step, serializable_args,
                                                              pbar_info_copy), error_callback=err_call))
        graph_results_n = [task.get() for task in build_graph_tasks_n]
        graph_results_b = [task.get() for task in build_graph_tasks_b]

        # 2.图合并
        build_graph_results_n = GraphMerger(graph_results_n, parallel_params[0], pbar_info=pbar_info).merge_graph()
        build_graph_results_b = GraphMerger(graph_results_b, parallel_params[1], True,
                                            pbar_info=pbar_info).merge_graph()

        if len(build_graph_results_n) != len(build_graph_results_b):
            raise RuntimeError(f'Parallel merge failed because the dp of npu: {len(build_graph_results_n)} '
                               f'is inconsistent with that of bench: {len(build_graph_results_b)}!')
        serializable_args.rank_list = [result.rank for result in build_graph_results_n]
        if pbar_info:
            PbarInfo.del_progress_dict_item(pbar_info, list(set(ranks_n) | set(ranks_b)),
                                            [f'{Const.RANK}{result.rank}' for result in build_graph_results_n])
        # 3.并行图比对和输出
        export_res_task_list = []
        create_directory(args.output_path)
        for i, result_n in enumerate(build_graph_results_n):
            graph_n = result_n.graph
            graph_b = build_graph_results_b[i].graph
            graph_task_info = BuildGraphTaskInfo(
                _build_graph_info(os.path.join(npu_path, f'rank{graph_n.root.rank}'), args, graph_n),
                _build_graph_info(os.path.join(bench_path, f'rank{graph_b.root.rank}'), args, graph_b),
                f'rank{graph_n.root.rank}', f'rank{graph_b.root.rank}', current_time)
            pbar_info_copy = PbarInfo.update_task_id(pbar_info, f'{Const.RANK}{result_n.rank}')
            export_res_task_list.append(pool.apply_async(_compare_and_export_graph,
                                                         args=(graph_task_info, input_param, serializable_args, step,
                                                               pbar_info_copy), error_callback=err_call))
        export_res_list = [res.get() for res in export_res_task_list]
        if any(export_res_list):
            failed_names = list(filter(lambda x: x, export_res_list))
            logger.error(f'Unable to export compare graph results: {", ".join(failed_names)}.')
        else:
            logger.info('Successfully exported compare graph results.')


def _graph_service_parser(parser):
    # -------------------------- 基础必填参数 --------------------------
    parser.add_argument("-tp", "--target_path", dest="target_path", type=str,
                        help="<Required> The target path.", required=True)
    parser.add_argument("-o", "--output_path", dest="output_path", type=str,
                        help="<Required> The visualization task result out path.", required=True)
    # -------------------------- 基础可选参数 --------------------------
    parser.add_argument("-gp", "--golden_path", dest="golden_path", type=str,
                        help="<Optional> The golden path.", required=False)
    parser.add_argument("-lm", "--layer_mapping", dest="layer_mapping", type=str, nargs='?', const=True,
                        help="<Optional> The layer mapping file path.", required=False)
    parser.add_argument("-oc", "--overflow_check", dest="overflow_check", action="store_true",
                        help="<Optional> whether open overflow_check for graph.", required=False)
    parser.add_argument("-fm", "--fuzzy_match", dest="fuzzy_match", action="store_true",
                        help="<Optional> whether to perform a fuzzy match on the api name.", required=False)
    parser.add_argument("-tensor_log", "--is_print_compare_log", dest="is_print_compare_log", action="store_true",
                        help="<Optional> whether print tensor compare log for visualization task.", required=False)
    parser.add_argument("-progress_log", "--is_print_progress_log", dest="is_print_progress_log", action="store_true",
                        help="<Optional> whether print progress log for visualization task.", required=False)

    # -------------------------- 不同并行切分策略合并可选参数 --------------------------
    group_n = parser.add_argument_group("Parallel Parameters, "
                                        "used for graph merging under different parallel partitioning strategies")

    group_n.add_argument("--rank_size", type=int, nargs='+', help="<Optional> The rank size of dump path.",
                         required=False)
    group_n.add_argument("--tp", type=int, nargs='+',
                         help="<Optional, but required if rank_size is not empty> The tp size of dump path.",
                         required=False)
    group_n.add_argument("--pp", type=int, nargs='+',
                         help="<Optional, but required if rank_size is not empty> The pp size of dump path.",
                         required=False)
    group_n.add_argument("--vpp", type=int, nargs='+', default=[1], help="<Optional> The vpp size of dump path.",
                         required=False)
    group_n.add_argument("--order", type=str, nargs='+', default=['tp-cp-ep-dp-pp'],
                         help="<Optional> The order of dump path.", required=False)


def _graph_service_command(args):
    try:
        if args.is_print_progress_log:
            # 往ProgressInfo中记录error日志，用于前端展示
            BaseLogger.error = get_log_msg_wrapper(BaseLogger.error)
        npu_path = args.target_path
        bench_path = args.golden_path
        ProgressInfo.print_progress_log = args.is_print_progress_log
        args.parallel_merge = True if args.rank_size else False
        args.parallel_params = load_parallel_param(args) if args.parallel_merge else None
        check_file_or_directory_path(npu_path, isdir=True)
        if bench_path:
            check_file_or_directory_path(bench_path, isdir=True)
        if check_file_type(npu_path) == FileCheckConst.DIR and not bench_path:
            content = check_directory_content(npu_path)
            if content == GraphConst.RANKS:
                _build_graph_ranks_with_pbar(args)
            elif content == GraphConst.STEPS:
                _build_graph_steps_with_pbar(args)
            else:
                _build_graph_with_pbar(npu_path, args)
        elif check_file_type(npu_path) == FileCheckConst.DIR and check_file_type(bench_path) == FileCheckConst.DIR:
            content_n = check_directory_content(npu_path)
            content_b = check_directory_content(bench_path)
            if content_n != content_b:
                raise ValueError('The directory structures of npu_path and bench_path are inconsistent.')
            input_param = {
                'npu_path': args.target_path,
                'bench_path': args.golden_path,
                'is_print_compare_log': args.is_print_compare_log
            }
            if content_n == GraphConst.RANKS:
                _compare_graph_ranks_with_pbar(input_param, args)
            elif content_n == GraphConst.STEPS:
                _compare_graph_steps_with_pbar(input_param, args)
            else:
                _compare_graph_with_pbar(input_param, args)
        else:
            logger.error("The npu_path or bench_path should be a folder.")
            raise CompareException(CompareException.INVALID_COMPARE_MODE)
    except KeyboardInterrupt:
        logger.warning("Interrupted by user, terminating processes and cleaning up...")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        raise e
    finally:
        ProgressInfo.update_process_running(False)


@dataclass
class ProgressConfig:
    core_func: Callable
    get_ranks: Callable
    db_name: str
    pbar_info_kwargs: dict = None
    use_monitor_thread: bool = True
    tqdm_total: Optional[int] = None


def _run_with_progress(param, args, config: ProgressConfig):
    """通用进度条处理"""

    monitor_thread = None
    pbar_info = None
    ranks = None

    try:
        if config.use_monitor_thread:
            manager = Manager()
            progress_dict = manager.dict()
            pbar_info = PbarInfo(progress_dict=progress_dict, **config.pbar_info_kwargs)
            ranks = config.get_ranks(args)
        else:
            pbar_info = PbarInfo(**config.pbar_info_kwargs)

        tqdm_args = {
            "desc": GraphConst.PBAR_DESC_PREFIX,
            "total": config.tqdm_total if config.tqdm_total is not None else pbar_info.total,
            "bar_format": GraphConst.BAR_FORMAT
        }

        with tqdm(**tqdm_args) as pbar:
            # 单进程场景直接更新pbar，多进程场景需要通过monitor thread从共享dict中获取进度更新pbar
            if config.use_monitor_thread:
                monitor_thread = threading.Thread(target=monitor_progress,
                                                  args=(pbar_info, pbar, ranks, args.parallel_merge))
                monitor_thread.start()
            else:
                pbar_info.pbar = pbar

            if param:
                config.core_func(param, args, pbar_info=pbar_info)
            else:
                config.core_func(args, pbar_info=pbar_info)

            db_path = os.path.join(args.output_path, config.db_name)
            post_process_db(db_path, pbar_info=pbar_info, is_parallel_merge=args.parallel_merge)

            if not args.parallel_merge and config.use_monitor_thread:
                distributed_analyse(db_path, args.overflow_check, pbar_info=pbar_info)

            if config.use_monitor_thread and monitor_thread:
                monitor_thread.join(timeout=5)

    except KeyboardInterrupt:
        logger.warning("Interrupted by user, terminating processes and cleaning up...")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        raise e
    finally:
        ProgressInfo.update_process_running(False)
        if config.use_monitor_thread and pbar_info:
            pbar_info.stop_monitor = True


def _build_graph_ranks_with_pbar(args):
    def core_func(args, pbar_info):
        if args.parallel_merge:
            _build_graph_ranks_parallel(args, pbar_info=pbar_info)
        else:
            _build_graph_ranks(args, pbar_info=pbar_info)

    def get_ranks(args):
        return check_and_return_dir_contents(args.target_path, Const.RANK)

    stage_total = _get_parallel_stage_total(args) if args.parallel_merge else GraphConst.BUILD_STAGES_TOTAL

    _run_with_progress(
        param=None,
        args=args,
        config=ProgressConfig(
            core_func=core_func,
            get_ranks=get_ranks,
            pbar_info_kwargs={"stage_total": stage_total},
            db_name=build_output_db_name,
        )
    )


def _build_graph_steps_with_pbar(args):
    steps = check_and_return_dir_contents(args.target_path, Const.STEP)

    def get_ranks(args):
        return check_and_return_dir_contents(os.path.join(args.target_path, steps[0]), Const.RANK)

    stage_total = _get_parallel_stage_total(args, steps) if args.parallel_merge else GraphConst.BUILD_STAGES_TOTAL

    _run_with_progress(
        param=None,
        args=args,
        config=ProgressConfig(
            core_func=_build_graph_steps,
            get_ranks=get_ranks,
            pbar_info_kwargs={"step_total": len(steps), "stage_total": stage_total},
            db_name=build_output_db_name,
        )
    )


def _build_graph_with_pbar(npu_path, args):
    def core_func(param, args, pbar_info):
        result = _build_graph_result(param, args, pbar_info)
        create_directory(args.output_path)
        file_name = _export_build_graph_result(args, result, pbar_info)
        if file_name:
            logger.error('Failed to export model build graph.')

    _run_with_progress(
        param=npu_path,
        args=args,
        config=ProgressConfig(
            core_func=core_func,
            get_ranks=lambda x: None,
            pbar_info_kwargs={},
            db_name=build_output_db_name,
            use_monitor_thread=False,
            tqdm_total=GraphConst.PBAR_TOTAL
        )
    )


def _compare_graph_ranks_with_pbar(input_param, args):
    def core_func(param, args, pbar_info):
        if args.parallel_merge:
            _compare_graph_ranks_parallel(param, args, pbar_info=pbar_info)
        else:
            _compare_graph_ranks(param, args, pbar_info=pbar_info)

    def get_ranks(args):
        if args.parallel_merge:
            return calculate_list(args.target_path, args.golden_path, mode=GraphConst.UNION)
        return calculate_list(args.target_path, args.golden_path)

    stage_total = _get_parallel_stage_total(args, is_compare=True) if args.parallel_merge \
        else GraphConst.COMPARE_STAGES_TOTAL

    _run_with_progress(
        param=input_param,
        args=args,
        config=ProgressConfig(
            core_func=core_func,
            get_ranks=get_ranks,
            pbar_info_kwargs={"stage_total": stage_total},
            db_name=compare_output_db_name
        )
    )


def _compare_graph_steps_with_pbar(input_param, args):
    steps = calculate_list(args.target_path, args.golden_path, Const.STEP)

    def get_ranks(args):
        rank_path_t = os.path.join(args.target_path, steps[0])
        rank_path_g = os.path.join(args.golden_path, steps[0])
        if args.parallel_merge:
            return calculate_list(rank_path_t, rank_path_g, mode=GraphConst.UNION)
        return calculate_list(rank_path_t, rank_path_g)

    stage_total = _get_parallel_stage_total(args, steps, is_compare=True) if args.parallel_merge \
        else GraphConst.COMPARE_STAGES_TOTAL

    _run_with_progress(
        param=input_param,
        args=args,
        config=ProgressConfig(
            core_func=_compare_graph_steps,
            get_ranks=get_ranks,
            pbar_info_kwargs={"stage_total": stage_total, "step_total": len(steps)},
            db_name=compare_output_db_name
        )
    )


def _compare_graph_with_pbar(input_param, args):
    def core_func(param, args, pbar_info):
        result = _compare_graph_result(param, args, pbar_info=pbar_info)
        create_directory(args.output_path)
        file_name = _export_compare_graph_result(args, result, pbar_info=pbar_info)
        if file_name:
            logger.error('Failed to export model compare graph.')

    _run_with_progress(
        param=input_param,
        args=args,
        config=ProgressConfig(
            core_func=core_func,
            get_ranks=lambda x: None,
            pbar_info_kwargs={"pbar": None, "stage_total": GraphConst.COMPARE_STAGES_TOTAL},
            db_name=compare_output_db_name,
            use_monitor_thread=False,
            tqdm_total=GraphConst.PBAR_TOTAL
        )
    )


def _get_parallel_stage_total(args, steps=None, is_compare=False):
    """
    获取不同并行切分策略的任务阶段数
    """
    parallel_params = args.parallel_params
    if not is_compare and (not parallel_params or len(parallel_params) != 1):
        raise RuntimeError('Parallel params error in build graph!')
    if is_compare and (not parallel_params or len(parallel_params) != 2):
        raise RuntimeError('Parallel params error in compare graph!')
    target_path = os.path.join(args.target_path, steps[0]) if steps else args.target_path
    validate_parallel_param(parallel_params[0], target_path)

    if is_compare:
        golden_path = os.path.join(args.golden_path, steps[0]) if steps else args.golden_path
        validate_parallel_param(parallel_params[1], golden_path, '[Bench]')

    stage_count_map = {
        "TPMerger": lambda param: param.rank_size // param.tp,
        "PPMerger": lambda param: param.rank_size // param.pp,
        "VPPMerger": lambda param: param.rank_size // param.pp,
        "TPPPMerger": lambda param: param.rank_size // param.pp + param.rank_size // param.pp // param.tp,
        "FullMerger": lambda param: param.rank_size // param.pp + param.rank_size // param.pp // param.tp,
        "NoParallelMerger": 0
    }

    def _get_stage_count(parallel_param, merger_name: str) -> int:
        rule = stage_count_map.get(merger_name, 0)
        return rule(parallel_param) if callable(rule) else rule

    merger_name_t = GraphMerger([], parallel_params[0]).strategy.__class__.__name__
    stage_count_target = _get_stage_count(parallel_params[0], merger_name_t)

    if is_compare:
        merger_name_g = GraphMerger([], parallel_params[1]).strategy.__class__.__name__
        stage_count_golden = _get_stage_count(parallel_params[1], merger_name_g)
        return GraphConst.COMPARE_STAGES_TOTAL + stage_count_target + stage_count_golden

    return GraphConst.BUILD_STAGES_TOTAL + stage_count_target


class CompareGraphResult:
    def __init__(self, graph_n, graph_b, graph_comparator, micro_steps, rank=0, step=0):
        self.graph_n = graph_n
        self.graph_b = graph_b
        self.graph_comparator = graph_comparator
        self.micro_steps = micro_steps
        self.rank = rank
        self.step = step


class BuildGraphResult:
    def __init__(self, graph, micro_steps=0, rank=0, step=0):
        self.graph = graph
        self.micro_steps = micro_steps
        self.rank = rank
        self.step = step


class PbarInfo:
    def __init__(self, pbar=None, progress_dict=None, task_id=None, step=0, step_total=1,
                 stage_total=GraphConst.BUILD_STAGES_TOTAL):
        self.pbar = pbar
        self.progress_dict = progress_dict
        self.task_id = task_id
        self.step = step
        self.step_total = step_total
        self.total = GraphConst.PBAR_TOTAL * step_total
        self.stage_total = stage_total * step_total  # 有几个阶段
        self.current_stage_dict = Manager().dict()  # 当前阶段，进程共享
        self.stage_progress = round(self.total / self.stage_total, 2)  # 每个阶段的最大进度
        self.stop_monitor = False
        self.wait_monitor = False
        self.continue_monitor = True

    def __deepcopy__(self, memo):
        new_obj = PbarInfo()
        new_obj.progress_dict = self.progress_dict
        new_obj.task_id = self.task_id
        new_obj.step = self.step
        new_obj.step_total = self.step_total
        new_obj.stage_total = self.stage_total
        new_obj.current_stage_dict = self.current_stage_dict
        new_obj.stage_progress = self.stage_progress
        new_obj.total = self.total
        new_obj.stop_monitor = self.stop_monitor
        new_obj.wait_monitor = self.wait_monitor
        new_obj.continue_monitor = self.continue_monitor
        return new_obj

    @staticmethod
    def update_task_id(pbar_info, task_id):
        """
        在进程池中，实例作为入参，修改实例属性，需要深拷贝实例使修改生效
        """
        if pbar_info:
            pbar_info.task_id = task_id
            return deepcopy(pbar_info)
        return pbar_info

    @staticmethod
    def del_progress_dict_item(pbar_info, origin_ranks, merged_ranks):
        """
        不同并行切分策略的图合并场景下，graph合并到一些rank中，剩余的rank作为task_id不再需要
        """
        diff_ranks = list(set(origin_ranks) - set(merged_ranks))
        for rank in diff_ranks:
            if rank in pbar_info.progress_dict:
                del pbar_info.progress_dict[rank]

    @staticmethod
    def reset_progress_and_current_stage(pbar_info, task_ids):
        """
        不同并行切分策略的图合并场景下，每个step需要重置进度信息
        """
        for task_id in task_ids:
            pbar_info.progress_dict[task_id] = GraphConst.PBAR_TOTAL * pbar_info.step
            pbar_info.current_stage_dict[task_id] = pbar_info.stage_total // pbar_info.step_total * pbar_info.step

    def set_continue_monitor(self, value: bool):
        self.continue_monitor = value
        self.wait_monitor = not value

    def set_wait_monitor(self, value: bool):
        self.wait_monitor = value
        self.continue_monitor = not value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import namedtuple
from msprobe.core.common.file_utils import load_yaml, check_file_or_directory_path
from msprobe.core.common.utils import is_int
from msprobe.pytorch.dump.pt_config import ACCCheckConfig


AccCheckConfig = namedtuple('AccCheckConfig', ['forward_content', 'backward_content',
                                               'result_csv_path', 'details_csv_path',
                                         'save_error_data', 'is_continue_acc_check', 'real_data_path', 'white_list',
                                         'black_list', 'error_data_path'])


class Config:
    def __init__(self, yaml_file):
        check_file_or_directory_path(yaml_file, False)
        config = load_yaml(yaml_file)
        self.config = {key: self.validate(key, value) for key, value in config.items()}

    def __getattr__(self, item):
        return self.config[item]

    def __str__(self):
        return '\n'.join(f"{key}={value}" for key, value in self.config.items())

    @staticmethod
    def validate(key, value):
        validators = {
            'white_list': list,
            'black_list': list,
            'error_data_path': str,
            'precision': int
        }
        if key not in validators:
            raise ValueError(f"{key} must be one of {validators.keys()}")
        if not isinstance(value, validators.get(key)):
            raise ValueError(f"{key} must be {validators[key].__name__} type")
        if key == 'precision' and not is_int(value):
            raise ValueError("precision must be an integer")
        if key == 'precision' and (value < 0 or value > 20):
            raise ValueError("precision must be greater than or equal to 0 and less than 21")
        if key == 'white_list':
            ACCCheckConfig.check_filter_list_config(key, value)
        if key == 'black_list':
            ACCCheckConfig.check_filter_list_config(key, value)
        if key == 'error_data_path':
            ACCCheckConfig.check_error_data_path_config(value)
        return value


cur_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
yaml_path = os.path.join(cur_path, "config.yaml")
msCheckerConfig = Config(yaml_path)


class CheckerConfig:
    def __init__(self, task_config=None):
        self.white_list = msCheckerConfig.white_list
        self.black_list = msCheckerConfig.black_list
        self.error_data_path = msCheckerConfig.error_data_path

        if task_config:
            self.load_config(task_config)

    def load_config(self, task_config):
        self.white_list = task_config.white_list
        self.black_list = task_config.black_list
        self.error_data_path = task_config.error_data_path
    

    def get_acc_check_config(self, **config_params):
        return AccCheckConfig(
            forward_content=config_params.get('forward_content'),
            backward_content=config_params.get('backward_content'),
            result_csv_path=config_params.get('result_csv_path'),
            details_csv_path=config_params.get('details_csv_path'),
            save_error_data=config_params.get('save_error_data'),
            is_continue_acc_check=config_params.get('is_continue_acc_check'),
            real_data_path=config_params.get('real_data_path'),
            white_list=self.white_list.copy() if self.white_list else [],
            black_list=self.black_list.copy() if self.black_list else [],
            error_data_path=config_params.get('error_data_path')
        )



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
from dataclasses import dataclass
from collections import defaultdict

import numpy as np
import pandas as pd
from tqdm import tqdm

from msprobe.core.common.const import CompareConst, Const
from msprobe.core.common.exceptions import FileCheckException
from msprobe.core.common.file_utils import load_json, remove_path, create_directory, save_excel, save_json
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException, add_time_with_xlsx, check_op_str_pattern_valid, \
    set_dump_path, get_dump_mode, check_compare_param, load_stack_json, get_file_type, add_time_with_json
from msprobe.core.compare.check import check_dump_json_str, check_stack_json_str, cross_dtype_mapping, \
    check_configuration_param
from msprobe.core.compare.utils import print_compare_ends_info, read_op, set_stack_json_path, reorder_index
from msprobe.core.compare.config import ModeConfig, MappingConfig, MappingDict
from msprobe.core.compare.multiprocessing_compute import CompareRealData
from msprobe.core.compare.diff_analyze.first_diff_analyze import FirstDiffAnalyze
from msprobe.core.compare.indicator_analysis.calculator import calculate_excel_result_df
from msprobe.core.compare.stats_diff_calc import ValType, ALL_TYPES


@dataclass
class ComparisonConfig:
    dump_mode: str
    stack_mode: bool
    fuzzy_match: bool
    data_mapping: dict
    suffix: str
    cell_mapping: dict
    api_mapping: dict
    layer_mapping: dict
    compared_file_type: str
    first_diff_analyze: bool
    is_print_compare_log: bool


class Comparator:
    def __init__(self, file_reader, mode_config: ModeConfig, mapping_config: MappingConfig, is_cross_framework=False):
        self.file_reader = file_reader
        self.mode_config = mode_config
        self.mapping_config = mapping_config
        self.cross_frame = is_cross_framework
        self.mapping_dict = MappingDict(mapping_config)

    def process_output_file(self, output_path, suffix, compared_file_type):
        file_name_prefix_mapping = {
            Const.DUMP_JSON_FILE: "compare_result",
            Const.DEBUG_JSON_FILE: "debug_compare_result"
        }
        file_name_prefix = file_name_prefix_mapping.get(compared_file_type, "compare_result")
        if self.mode_config.first_diff_analyze:
            file_name = add_time_with_json("compare_result" + suffix)
        else:
            file_name = add_time_with_xlsx(file_name_prefix + suffix)
        file_path = os.path.join(os.path.realpath(output_path), file_name)
        if os.path.exists(file_path):
            logger.warning(f"{file_path} will be deleted.")
            remove_path(file_path)
        return file_path

    def compare_core(self, input_param, output_path, **kwargs):
        """
        Compares data from multiple JSON files and generates a comparison report.

        Args:
            input_param (dict): A dictionary containing paths to JSON files ("npu_path", "bench_path",
                                "stack_path").
            output_path (str): The path where the output Excel report will be saved.
            **kwargs: Additional keyword arguments including:
            - stack_mode (bool, optional): Enables stack mode comparison. Defaults to False.
            - auto_analyze (bool, optional): If True, triggers automatic analysis after comparison. Defaults to True.
            - suffix (str, optional): Suffix to append to the output file name. Defaults to ''.
            - fuzzy_match (bool, optional): Enables fuzzy matching during comparison. Defaults to False.
            - dump_mode (str): ALL, SUMMARY, MD5.

        Returns:
        """
        logger.info("Please check whether the input data belongs to you. If not, there may be security risks.")

        # get kwargs or set default value
        suffix = kwargs.get('suffix', '')
        rank = suffix[1:]

        # process output file
        file_path = self.process_output_file(output_path, suffix, self.mode_config.compared_file_type)

        # initialize the compare result table and compare general data(name, dtype, shape, statistics/md5, etc.)
        npu_json = input_param.get("npu_path")
        bench_json = input_param.get("bench_path")
        stack_json = input_param.get("stack_path")
        parse_data = ParseData(self.mode_config, rank)  # load and parse json data
        npu_df, bench_df = parse_data.parse([npu_json, bench_json, stack_json])

        logger.info("Matching APIs/Modules in progress...")
        result_df = self.compare_statistics(npu_df, bench_df)
        logger.info("APIs/Modules match done.")
        if result_df.empty:
            logger.warning("Can`t match any op. No compare result file generated.")
            return

        if self.mode_config.first_diff_analyze:
            # add P2POp additional info from npu_df and bench_df to result_df
            result_df['NPU P2POp op'] = npu_df['op']
            result_df['Bench P2POp op'] = bench_df['op']
            result_df['NPU P2POp peer'] = npu_df['peer']
            result_df['Bench P2POp peer'] = bench_df['peer']

            first_diff_analyze = FirstDiffAnalyze(self.mode_config, rank)
            check_result = first_diff_analyze.check(result_df)
            save_json(file_path, check_result, indent=4)
            logger.info(f"Saving json file to disk: {file_path}")
            return

        # compare real data
        if self.mode_config.dump_mode == Const.ALL:
            logger.info("Compare real data in progress...")
            compare_real_data = CompareRealData(self.file_reader, self.mode_config, self.cross_frame)
            result_df = compare_real_data.do_multi_process(input_param, result_df)
            logger.info("Compare real data done.")

        # calculate Indicators
        logger.info("Calculating comparison indicators in progress...")
        calculate_excel_result_df(result_df, self.mode_config.dump_mode)
        result_df.drop(columns=['state', 'api_origin_name'], inplace=True)  # 删除中间数据，两列不落盘
        logger.info("Comparison indicators calculation done.")

        # save result excel file
        logger.info("Saving result excel file in progress...")
        logger.info(f"The result excel file path is: {file_path}.")
        save_excel(file_path, result_df)

        print_compare_ends_info()

    def compare_statistics(self, npu_df, bench_df):
        npu_df[[Const.DTYPE, Const.SHAPE]] = npu_df[[Const.DTYPE, Const.SHAPE]].astype(str)
        bench_df[[Const.DTYPE, Const.SHAPE]] = bench_df[[Const.DTYPE, Const.SHAPE]].astype(str)

        # create new columns for compare op_name and shape
        # process npu_df's COMPARE_KEY whether same or different framework
        process_df = ProcessDf(self.mode_config, self.mapping_config, self.mapping_dict)
        # 处理重计算对应的backward的序号。反向重计算序号调整属于精确匹配，与模糊匹配互斥。
        if not self.mode_config.fuzzy_match:
            npu_df = process_df.update_backward_call(npu_df)
            bench_df = process_df.update_backward_call(bench_df)
        npu_df, bench_df = process_df.process_compare_key_and_shape(npu_df, bench_df)

        # match npu and bench, match_result contains both npu_info and bench_info
        match = Match(self.mode_config, self.mapping_config, self.cross_frame)
        match_result = match.match_api_infos(npu_df, bench_df)
        # 筛选出npu_name存在的行并填充筛选出行中的缺失值为N/A
        match_result = match_result[match_result['op_name_x'].notna()].fillna(CompareConst.N_A)
        bench_columns = [i + '_y' for i in bench_df.columns]
        match_result.loc[~match.gen_dtype_condition(match_result), bench_columns] = CompareConst.N_A

        # organize compare result table by renaming columns
        if self.mode_config.dump_mode == Const.ALL and self.mode_config.first_diff_analyze:
            self.mode_config.dump_mode = Const.SUMMARY
        create_table = CreateTable(self.mode_config)
        result_df, header = create_table.make_result_df(match_result)

        # calculate statistics diff
        calc_stats_diff = CalcStatsDiff(self.mode_config)
        return calc_stats_diff.calc_accuracy(result_df, header)


class ParseData:
    def __init__(self, mode_config: ModeConfig, rank):
        self.mode_config = mode_config
        self.rank = rank

    @staticmethod
    def get_direction_and_call_direction(data_name):
        parts = data_name.split(Const.SEP)
        if len(parts) < 2:
            return None, None

        last = parts[-1]
        second_last = parts[-2]

        # direction 判断
        if Const.BACKWARD in (last, second_last):
            direction = Const.BACKWARD
        elif Const.FORWARD in (last, second_last):
            direction = Const.FORWARD
        else:
            return None, None

        # call_direction 永远取最后两个
        call_direction = Const.SEP.join((second_last, last))

        return direction, call_direction

    @staticmethod
    def get_op_no_number(data_name):
        parts = data_name.split(Const.SEP)
        return Const.SEP.join(parts[:-2]) if len(parts) > 2 else ''

    def parse(self, file_list):
        npu_json_path, bench_json_path, stack_json_path = file_list
        npu_json_data = load_json(npu_json_path)
        bench_json_data = load_json(bench_json_path)
        stack_json_data = load_stack_json(stack_json_path) if self.mode_config.stack_mode else None

        # parse json data and generate df
        npu_df = self.gen_data_df(npu_json_data, stack_json_data, 'NPU')
        bench_df = self.gen_data_df(bench_json_data, stack_json_data, 'Bench')

        return npu_df, bench_df

    def init_result(self):
        result = {
            CompareConst.OP_NAME: [],
            Const.DTYPE: [],
            Const.SHAPE: [],
            Const.SUMMARY: [],
            Const.STACK_INFO: [],
            Const.STATE: [],
            Const.API_ORIGIN_NAME: [],
            Const.REQ_GRAD: [],
            Const.DIRECTION: [],  # 目前三种选择：'forward', 'backward', None
            Const.CALL_DIRECTION: [],
            Const.OP_NO_NUMBER: [],  # 删除调用序号
            Const.BACKWARD_CALL_ORDER: [],  # 计算op反向调用顺序，初始化为0
            Const.SUFFIX: []
        }
        if self.mode_config.dump_mode == Const.ALL:
            result[Const.DATA_NAME] = []
        elif self.mode_config.dump_mode == Const.MD5:
            result[Const.MD5] = []
        return result

    def create_progress_bar(self, api_nums: int, device: str):
        desc = f'{device} API/Module Read Progress'
        if self.rank:
            desc = f'[{self.rank}]' + desc
        return tqdm(total=api_nums, desc=desc, unit="api/module", ncols=100)

    def gen_data_df(self, data_json, stack_json_data, device: str):
        result = self.init_result()
        op_backward_count = defaultdict(int)  # 记录backward调用次数，用于后续调用序更新

        apis_data = data_json.get('data', None)
        if not apis_data:
            logger.warning('No APIs found in dump.json.')
            return pd.DataFrame(result)

        progress_bar = self.create_progress_bar(len(apis_data), device)

        # 从json中循环解析API数据，遍历所有API
        for data_name in apis_data:
            check_op_str_pattern_valid(data_name)

            op_parsed_list = self.gen_merge_list(data_json, data_name, stack_json_data)
            if not op_parsed_list:
                progress_bar.update(1)
                continue

            op_no_number = self.get_op_no_number(data_name)
            if op_no_number == '':
                progress_bar.update(1)
                continue

            direction, call_direction = self.get_direction_and_call_direction(data_name)
            backward_call_order = op_backward_count.get(op_no_number, 0)
            if direction == Const.BACKWARD:
                op_backward_count[op_no_number] += 1

            reordered_index_list = reorder_index(op_parsed_list)

            stack_info = op_parsed_list[-1].get('full_info') if self.mode_config.stack_mode else None

            for i, index in enumerate(reordered_index_list):
                op_item = op_parsed_list[index]
                summary_data = [
                    str(op_item.get(key)) if op_item.get(key) is None else op_item.get(key)
                    for key in Const.SUMMARY_METRICS_LIST
                ]
                full_op_name = op_item.get(Const.FULL_OP_NAME)
                suffix = full_op_name.replace(data_name, '')

                # common key
                result[CompareConst.OP_NAME].append(full_op_name)
                result[Const.DTYPE].append(op_item.get(Const.DTYPE))
                result[Const.SHAPE].append(op_item.get(Const.SHAPE))
                result[Const.STATE].append(op_item.get(Const.STATE))
                result[Const.REQ_GRAD].append(op_item.get(Const.REQ_GRAD))
                result[Const.API_ORIGIN_NAME].append(data_name)
                result[Const.SUMMARY].append(summary_data)
                result[Const.DIRECTION].append(direction)
                result[Const.CALL_DIRECTION].append(call_direction)
                result[Const.OP_NO_NUMBER].append(op_no_number)
                result[Const.BACKWARD_CALL_ORDER].append(backward_call_order)
                result[Const.SUFFIX].append(suffix)

                # dump_mode differ key
                if self.mode_config.dump_mode == Const.MD5:
                    result[Const.MD5].append(op_item.get(Const.MD5))
                if self.mode_config.dump_mode == Const.ALL:
                    result[Const.DATA_NAME].append(op_item.get(Const.DATA_NAME))

                # mode_config stack_mode addition key
                result[Const.STACK_INFO].append(stack_info if i == 0 else None)

                # mode_config first_diff_analyze addition key
                if self.mode_config.first_diff_analyze:
                    result.setdefault('op', []).append(op_item.get('op', str(None)))
                    result.setdefault('peer', []).append(op_item.get('peer', str(None)))

            progress_bar.update(1)
        progress_bar.close()
        return pd.DataFrame(result)

    def gen_merge_list(self, json_data, op_name, stack_json_data):
        op_data = json_data['data'][op_name]
        if self.mode_config.compared_file_type == Const.DUMP_JSON_FILE:
            check_dump_json_str(op_data, op_name)
        op_parsed_list = read_op(op_data, op_name)

        if self.mode_config.stack_mode:
            stack_info = stack_json_data.get(op_name)
            if stack_info is not None:
                check_stack_json_str(stack_info, op_name)
        else:
            stack_info = None
        # always add stack_info whether stack_mode is True
        op_parsed_list.append({
            'full_op_name': op_name,
            'full_info': stack_info
        })
        return op_parsed_list


class ProcessDf:
    def __init__(self, mode_config: ModeConfig, mapping_config: MappingConfig, mapping_dict: MappingDict):
        self.mode_config = mode_config
        self.mapping_config = mapping_config
        self.mapping_dict = mapping_dict

    @staticmethod
    def get_api_name(api_list):
        try:
            api_name = api_list[0] + Const.SEP + api_list[1]
        except IndexError as error:
            logger.error('Failed to retrieve API name, please check if the dump data is reasonable')
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR) from error
        return api_name

    @staticmethod
    def update_backward_call(cmp_df):
        # ===============================
        # Step 1: 拆 call_direction 的两部分
        # ===============================
        cd_parts = cmp_df[Const.CALL_DIRECTION].str.rsplit(Const.SEP, n=1, expand=True)
        cd_parts.columns = ["cd_head", "cd_tail"]

        head_is_digit = cd_parts["cd_head"].str.isdigit()
        tail_is_digit = cd_parts["cd_tail"].str.isdigit()
        is_backward = cmp_df[Const.DIRECTION] == Const.BACKWARD

        # ===============================
        # Step 2: 矢量化替换 backward_call_order
        # ===============================
        mask_head = is_backward & head_is_digit
        mask_tail = is_backward & tail_is_digit
        update_mask = mask_head | mask_tail

        cd_parts.loc[mask_head, "cd_head"] = (
            cmp_df.loc[mask_head, Const.BACKWARD_CALL_ORDER].astype(str)
        )
        cd_parts.loc[mask_tail, "cd_tail"] = (
            cmp_df.loc[mask_tail, Const.BACKWARD_CALL_ORDER].astype(str)
        )

        # 只更新需要更新的 call_direction
        cmp_df.loc[update_mask, Const.CALL_DIRECTION] = (
                cd_parts.loc[update_mask, "cd_head"]
                + Const.SEP
                + cd_parts.loc[update_mask, "cd_tail"]
        )

        # ===============================
        # Step 3: 初始化 + 局部更新 OP_NAME_UPDATE
        # ===============================
        # 默认保持原值
        cmp_df[CompareConst.OP_NAME_UPDATE] = cmp_df[CompareConst.OP_NAME]

        # 只更新命中的行
        cmp_df.loc[update_mask, CompareConst.OP_NAME_UPDATE] = (
                cmp_df.loc[update_mask, Const.OP_NO_NUMBER]
                + Const.SEP
                + cmp_df.loc[update_mask, Const.CALL_DIRECTION]
                + cmp_df.loc[update_mask, Const.SUFFIX]
        )

        return cmp_df

    def process_compare_key_and_shape(self, npu_df, bench_df):
        npu_df = self.assign_npu_df_compare_key(npu_df, bench_df)
        npu_df[CompareConst.CMP_SHAPE] = npu_df[Const.SHAPE]
        bench_cmp_key = (
            bench_df[CompareConst.OP_NAME_UPDATE]
            if not self.mode_config.fuzzy_match
            else bench_df[CompareConst.OP_NAME]
        )
        bench_df[CompareConst.CMP_KEY] = bench_cmp_key
        bench_df[CompareConst.CMP_SHAPE] = bench_df[Const.SHAPE]
        return npu_df, bench_df

    def assign_npu_df_compare_key(self, npu_df, bench_df):
        """
        处理 npu_df 的 COMPARE_KEY 赋值逻辑

        :param npu_df: DataFrame，NPU 对比数据
        :param bench_df: DataFrame，Bench 对比数据
        :return: compare_key(name)处理后的 npu_df
        """
        # 处理api_mapping映射
        if self.mapping_config.api_mapping:
            # 如果用户不传api_mapping.yaml，先使用内置api_mapping.yaml替换npu_op_name
            npu_df[CompareConst.CMP_KEY] = npu_df[CompareConst.OP_NAME].apply(self.process_internal_api_mapping)
            # 如果用户传入api_mapping.yaml，再使用传入api_mapping.yaml进一步替换npu_op_name
            if isinstance(self.mapping_config.api_mapping, str):
                self.modify_compare_data_with_user_mapping(npu_df, bench_df)
        # 处理cell_mapping映射
        elif self.mapping_config.cell_mapping:
            npu_df[CompareConst.CMP_KEY] = npu_df[CompareConst.OP_NAME].apply(self.process_cell_mapping)
        # 处理data_mapping映射
        elif self.mapping_config.data_mapping:
            npu_df[CompareConst.CMP_KEY] = npu_df[CompareConst.OP_NAME].apply(self.process_data_mapping)
        else:
            cmp_key = (
                npu_df[CompareConst.OP_NAME_UPDATE]
                if not self.mode_config.fuzzy_match
                else npu_df[CompareConst.OP_NAME]
            )
            npu_df[CompareConst.CMP_KEY] = cmp_key
        return npu_df

    def process_internal_api_mapping(self, npu_op_name):
        # get api name & class name from op_name
        ms_api_name = self.get_api_name(npu_op_name.split(Const.SEP))
        class_name = ms_api_name.split(Const.SEP)[0]
        if class_name == "Mint":
            return npu_op_name.replace("Mint", "Torch")
        elif class_name == "MintFunctional":
            return npu_op_name.replace("MintFunctional", "Functional")
        elif self.mapping_dict.ms_to_pt_mapping.get(ms_api_name):
            return npu_op_name.replace(ms_api_name, self.mapping_dict.ms_to_pt_mapping.get(ms_api_name))
        else:
            return npu_op_name

    def modify_compare_data_with_user_mapping(self, npu_df, bench_df):
        def remove_prefix(string, prefix):
            if string.startswith(prefix):
                return string[len(prefix):]
            return string

        def gen_input_compare_key(pattern, term):
            is_unmatched = True
            for i, prefix in enumerate(mapping_dict.get(f'ms_{term}')):
                if remove_prefix(op_name, api_origin_name + pattern) == str(prefix):
                    npu_df.loc[index, CompareConst.CMP_KEY] = (
                        op_name.replace(pattern + str(prefix), pattern + str(mapping_dict.get(f'pt_{term}')[i])))
                    is_unmatched = False
            return is_unmatched

        ms_api_indices_dict = self.get_api_indices_dict(npu_df)
        pt_api_indices_dict = self.get_api_indices_dict(bench_df)

        for mapping_dict in self.mapping_dict.api_mapping_dict:
            all_length_equal = True
            for k1, k2 in CompareConst.API_MAPPING_KEYS_TO_COMPARE:
                if len(mapping_dict.get(k1, [])) != len(mapping_dict.get(k2, [])):
                    all_length_equal = False
            if not all_length_equal:
                logger.warning('The user-defined mapping table is incorrect,\
                                make sure that the number of parameters is equal')
                continue

            ms_api, pt_api = mapping_dict.get('ms_api'), mapping_dict.get('pt_api')
            if ms_api not in ms_api_indices_dict or pt_api not in pt_api_indices_dict:
                continue
            for index in ms_api_indices_dict.get(ms_api):
                op_name = npu_df.loc[index, CompareConst.OP_NAME].replace(ms_api, pt_api, 1)
                state = npu_df.loc[index, Const.STATE]
                api_origin_name = npu_df.loc[index, Const.API_ORIGIN_NAME].replace(ms_api, pt_api, 1)
                if state == Const.INPUT:
                    is_abandoned = gen_input_compare_key(CompareConst.INPUT_PATTERN, 'args')
                elif state == Const.KWARGS:
                    is_abandoned = gen_input_compare_key(CompareConst.KWARGS_PATTERN, 'args')
                elif state == Const.OUTPUT:
                    is_abandoned = gen_input_compare_key(CompareConst.OUTPUT_PATTERN, 'output')
                elif state == Const.PARAMS:
                    is_abandoned = gen_input_compare_key(CompareConst.PARAMS_PATTERN, 'parameters')
                elif state == Const.PARAMS_GRAD:
                    is_abandoned = gen_input_compare_key(CompareConst.PARAMS_GRAD_PATTERN, 'parameters_grad')
                else:
                    logger.error(f'Excepted op_name: {op_name}')
                    raise CompareException(CompareException.INVALID_DATA_ERROR)
                if is_abandoned:
                    npu_df.loc[index, CompareConst.CMP_KEY] = op_name + 'abandoned'

    def get_api_indices_dict(self, op_name_df):
        """
        生成多个api对应的各自的所有的input、output等的index的键值对字典
        示例：
        {'Functional.conv2d': [0, 1, 2, 3],
        'Functional.batch_norm': [4, 5, 6, 7, 8]
        }
        """
        api_indices_dict = defaultdict(list)
        for op_index, name in enumerate(op_name_df[CompareConst.OP_NAME]):
            api_name = self.get_api_name(name.split(Const.SEP))
            api_indices_dict[api_name].append(op_index)
        return api_indices_dict

    def process_cell_mapping(self, npu_op_name):
        if not npu_op_name:
            return CompareConst.N_A
        param_grad_flag = Const.PARAMS_GRAD in npu_op_name.split(Const.SEP)
        if not param_grad_flag and not re.search(Const.REGEX_FORWARD_BACKWARD, npu_op_name):
            return CompareConst.N_A
        npu_op_name = npu_op_name.replace("Cell", "Module", 1)
        if self.mapping_dict.cell_mapping_dict:
            # get cell name & class name from op_name
            # Cell.fc1.Dense.forward.0.input.0
            # npu_op_name.split(Const.SEP, 1)[-1] 表示Module或Cell字段后面的部分
            cell_name = re.split(r'\.(?:forward|backward|parameters_grad)\.', npu_op_name.split(Const.SEP, 1)[-1])[0]

            # 1. 精确整段匹配（保持原行为）
            if cell_name in self.mapping_dict.cell_mapping_dict:
                new_cell_name = self.mapping_dict.cell_mapping_dict[cell_name]
            # 2. 兜底：简单字符串替换（不做任何限制）
            else:
                new_cell_name = cell_name
                for target_name, golden_name in self.mapping_dict.cell_mapping_dict.items():
                    if target_name in new_cell_name:
                        new_cell_name = new_cell_name.replace(target_name, golden_name, 1)
            # 3. 应用替换
            if new_cell_name != cell_name:
                npu_op_name = npu_op_name.replace(cell_name, new_cell_name, 1)

        return npu_op_name

    def process_data_mapping(self, npu_op_name):
        return self.mapping_dict.data_mapping_dict.get(npu_op_name, npu_op_name)


class Match:
    def __init__(self, mode_config: ModeConfig, mapping_config: MappingConfig, cross_frame):
        self.mode_config = mode_config
        self.mapping_config = mapping_config
        self.cross_frame = cross_frame

    @staticmethod
    def put_unmatched_in_table(match_result, npu_op_item):
        npu_columns = npu_op_item.index.tolist()[:-2]
        bench_columns = [name + '_y' for name in npu_columns]
        na_series = pd.Series([CompareConst.N_A] * len(bench_columns), index=bench_columns)
        new_result_item = pd.concat([npu_op_item, na_series]).to_frame().T
        new_result_item.columns = CompareConst.MATCH_RESULT_COLUMNS
        match_result = pd.concat([match_result, new_result_item])
        return match_result

    @staticmethod
    def put_matched_in_table(match_result, npu_op_item, bench_op_item):
        head_len = len(CompareConst.MATCH_RESULT_COLUMNS)
        new_result_item = pd.concat([npu_op_item, bench_op_item]).head(head_len).to_frame().T
        new_result_item.columns = CompareConst.MATCH_RESULT_COLUMNS
        match_result = pd.concat([match_result, new_result_item])
        return match_result

    @staticmethod
    def rename_api(op_name):
        """
        原api： {api_type}.{api_name}.{API调用次数}.{前向反向}.{input/output}.{参数序号}
        rename后： {api_type}.{api_name}.{前向反向}.{input/output}.{参数序号}
        """
        if Const.FORWARD not in op_name and Const.BACKWARD not in op_name:
            return op_name
        process = Const.FORWARD if Const.FORWARD in op_name else Const.BACKWARD
        name_split = op_name.split(process)
        try:
            torch_func_index, in_out = name_split[0], name_split[1]
        except IndexError as error:
            logger.error(f'{op_name} can not be split with {process}, please check!')
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR) from error
        torch_func_split = torch_func_index.rsplit(Const.SEP, 2)
        torch_func = str(torch_func_split[0]) + Const.SEP + process + str(in_out)
        return torch_func

    def check_op_item(self, npu_op_item, bench_op_item):
        name_match = self.rename_api(npu_op_item[CompareConst.CMP_KEY]) == self.rename_api(
            bench_op_item[CompareConst.CMP_KEY])
        if name_match:
            return True
        else:
            npu_op_name = npu_op_item[CompareConst.OP_NAME]
            bench_op_name = bench_op_item[CompareConst.OP_NAME]
            check_op_str_pattern_valid(npu_op_name)
            check_op_str_pattern_valid(bench_op_name)
            logger.warning(f"{npu_op_name} and {bench_op_name} can not fuzzy match")
            return False

    def match_api_infos(self, npu_df, bench_df):
        """
        正常匹配和模糊匹配
        """
        if self.mapping_config.data_mapping:
            match_result = pd.merge(npu_df, bench_df, on=[CompareConst.CMP_KEY], how='left')

            # reorder match_result by op_name of npu
            op_name_order = npu_df[CompareConst.OP_NAME].tolist()
            match_result[CompareConst.OP_NAME_X] = pd.Categorical(match_result[CompareConst.OP_NAME_X],
                                                                  categories=op_name_order, ordered=True)
            match_result = match_result.sort_values(CompareConst.OP_NAME_X).reset_index(drop=True)
            match_result[CompareConst.OP_NAME_X] = match_result[CompareConst.OP_NAME_X].astype('object')
        elif not self.mode_config.fuzzy_match:
            match_result = pd.merge(npu_df, bench_df, on=[CompareConst.CMP_KEY],
                                    how='outer')
        else:
            drop_list = [Const.DIRECTION, Const.CALL_DIRECTION, Const.OP_NO_NUMBER, Const.BACKWARD_CALL_ORDER,
                         Const.SUFFIX]
            npu_df.drop(columns=drop_list, inplace=True)
            bench_df.drop(columns=drop_list, inplace=True)
            match_result = self.process_fuzzy_match(npu_df, bench_df)
        return match_result

    def process_fuzzy_match(self, npu_df, bench_df):
        """
        模糊匹配通过循环方式匹配api
        """
        npu_ops_queue = []
        bench_ops_queue = []
        match_result = pd.DataFrame(columns=CompareConst.MATCH_RESULT_COLUMNS)

        max_len = max(len(npu_df), len(bench_df))
        min_len = min(len(npu_df), len(bench_df))
        for i in range(max_len):
            if i < min_len:
                npu_ops_queue.append(npu_df.iloc[i])
                bench_ops_queue.append(bench_df.iloc[i])
            else:
                try:
                    npu_ops_queue.append(npu_df.iloc[i])
                except IndexError:
                    pass
                try:
                    bench_ops_queue.append(bench_df.iloc[i])
                except IndexError:
                    pass

            # 如果append之后queue状态不一致，则判断结束
            if bool(npu_ops_queue) ^ bool(bench_ops_queue):
                break

            npu_match_point, bench_match_point = self.match_op(npu_ops_queue, bench_ops_queue)

            # 如果没有匹配到，数据放到队列中，跳过。直到后面匹配到，把匹配之前的api放到不匹配中
            if npu_match_point == -1 and bench_match_point == -1:
                continue

            npu_op_item = npu_ops_queue[npu_match_point]
            bench_op_item = bench_ops_queue[bench_match_point]
            unmatched_data = npu_ops_queue[0: npu_match_point]
            for op_item in unmatched_data:
                match_result = self.put_unmatched_in_table(match_result, op_item)
            match_result = self.put_matched_in_table(match_result, npu_op_item, bench_op_item)
            del npu_ops_queue[0: npu_match_point + 1]
            del bench_ops_queue[0: bench_match_point + 1]

        if npu_ops_queue:
            for op_item in npu_ops_queue:
                match_result = self.put_unmatched_in_table(match_result, op_item)

        match_result.reset_index(drop=True, inplace=True)
        return match_result

    def match_op(self, npu_queue, bench_queue):
        for b_index, b_op in enumerate(bench_queue[0: -1]):
            if self.check_op_item(npu_queue[-1], b_op):
                return len(npu_queue) - 1, b_index
        if self.check_op_item(npu_queue[-1], bench_queue[-1]):
            return len(npu_queue) - 1, len(bench_queue) - 1
        for n_index, n_op in enumerate(npu_queue[0: -1]):
            if self.check_op_item(n_op, bench_queue[-1]):
                return n_index, len(bench_queue) - 1
        return -1, -1

    def gen_dtype_condition(self, match_result):
        """
        dtype匹配条件为npu、bench的dtype一致或属于规定的映射关系
        """
        # 如果使用了data_mapping，不校验dtype，返回全True的DataFrame
        if self.mapping_config.data_mapping:
            return pd.Series(True, index=match_result.index)

        npu_dtype = match_result['dtype_x']
        bench_dtype = match_result['dtype_y']
        npu_dtype = self.process_cross_frame_dtype(npu_dtype)
        bench_dtype = self.process_cross_frame_dtype(bench_dtype)

        equal_condition = npu_dtype == bench_dtype
        match_condition = (
                (npu_dtype.isin(CompareConst.DTYPE_MATCH_GROUPS[0]) & bench_dtype.isin(
                    CompareConst.DTYPE_MATCH_GROUPS[0])) |
                (npu_dtype.isin(CompareConst.DTYPE_MATCH_GROUPS[1]) & bench_dtype.isin(
                    CompareConst.DTYPE_MATCH_GROUPS[1]))
        )
        return equal_condition | match_condition

    def process_cross_frame_dtype(self, dtype):
        if self.cross_frame:
            dtype = dtype.map(cross_dtype_mapping).fillna(dtype)
        return dtype


class CreateTable:
    def __init__(self, mode_config: ModeConfig):
        self.mode_config = mode_config

    @staticmethod
    def process_data_name(result):
        result['data_name_x'] = result.apply(lambda row: [row['data_name_x'], row['data_name_y']], axis=1)
        return result

    @staticmethod
    def set_summary(summary):
        if summary == CompareConst.N_A:
            return [CompareConst.N_A] * 4  # 4为统计值个数
        summary_list = []
        for i in summary:
            if str(i).lower() == 'nan':
                summary_list.append(CompareConst.NAN)
            else:
                summary_list.append(i)
        return summary_list

    def make_result_df(self, result):
        # get header
        header = CompareConst.HEAD_OF_COMPARE_MODE[self.mode_config.dump_mode][:]
        if self.mode_config.stack_mode:
            header.append(CompareConst.STACK)
        if self.mode_config.dump_mode == Const.ALL:
            header.append(CompareConst.DATA_NAME)
            result = self.process_data_name(result)

        # rename match_result columns
        result.rename(columns={'op_name_x': CompareConst.NPU_NAME,
                               'op_name_y': CompareConst.BENCH_NAME,
                               'dtype_x': CompareConst.NPU_DTYPE,
                               'dtype_y': CompareConst.BENCH_DTYPE,
                               'shape_x': CompareConst.NPU_SHAPE,
                               'shape_y': CompareConst.BENCH_SHAPE,
                               'md5_x': CompareConst.NPU_MD5,
                               'md5_y': CompareConst.BENCH_MD5,
                               'data_name_x': CompareConst.DATA_NAME,
                               'stack_info_x': CompareConst.STACK,
                               'state_x': Const.STATE,
                               'api_origin_name_x': Const.API_ORIGIN_NAME,
                               'requires_grad_x': CompareConst.NPU_REQ_GRAD,
                               'requires_grad_y': CompareConst.BENCH_REQ_GRAD
                               },
                      inplace=True)

        # process summary data
        npu_summary = [CompareConst.NPU_MAX, CompareConst.NPU_MIN, CompareConst.NPU_MEAN, CompareConst.NPU_NORM]
        bench_summary = [CompareConst.BENCH_MAX, CompareConst.BENCH_MIN, CompareConst.BENCH_MEAN,
                         CompareConst.BENCH_NORM]
        # process requires_grad
        result[CompareConst.REQ_GRAD_CONSIST] = result[CompareConst.NPU_REQ_GRAD] == result[CompareConst.BENCH_REQ_GRAD]

        if result.empty:
            result[npu_summary] = pd.DataFrame(columns=npu_summary)
            result[bench_summary] = pd.DataFrame(columns=bench_summary)
        else:
            result[npu_summary] = result['summary_x'].apply(self.set_summary).tolist()
            result[bench_summary] = result['summary_y'].apply(self.set_summary).tolist()

        header.extend([Const.STATE, Const.API_ORIGIN_NAME])
        result_df = pd.DataFrame(columns=header)
        for h in header:
            if h in result.columns:
                result_df[h] = result[h]
        return result_df, header


class CalcStatsDiff:
    def __init__(self, mode_config: ModeConfig):
        self.mode_config = mode_config
        self.rules = None
        self.build_rules()
        self.DEFAULT_RULE = self.static_diff(CompareConst.N_A)

    @staticmethod
    def is_same_value(a: pd.Series, b: pd.Series) -> pd.Series:
        """
        检查是否相等
        """
        return a.astype(str).eq(b.astype(str))

    @staticmethod
    def is_number(val: pd.Series) -> pd.Series:
        """
        检查是否为有效的数值，并排除布尔类型
        """
        val_str = val.astype(str)
        mask_bool = val_str.str.lower().eq('true') | val_str.str.lower().eq('false')
        mask_numeric = pd.to_numeric(val, errors='coerce').notna()
        return mask_numeric & ~mask_bool

    @staticmethod
    def is_nan(val: pd.Series) -> pd.Series:
        """
        检查是否为字符串形式的 'nan' (包括大小写) 或实际的 NaN
        """
        val_str = val.astype(str)
        return val.isna() | val_str.str.lower().eq('nan')

    @staticmethod
    def is_inf(val: pd.Series) -> pd.Series:
        """
        检查是否为正无穷 (inf)
        """
        val_str = val.astype(str)
        return (val == np.inf) | val_str.str.lower().eq('inf')

    @staticmethod
    def is_neg_inf(val: pd.Series) -> pd.Series:
        """
        检查是否为负无穷 (-inf)
        """
        val_str = val.astype(str)
        return (val == -np.inf) | val_str.str.lower().eq('-inf')

    @staticmethod
    def is_device(val: pd.Series) -> pd.Series:
        """
        检查是否包含 'npu', 'cpu' 或 'cuda' 字符串
        """
        return val.astype(str).str.contains('npu|cpu|cuda', case=False, na=False)

    @staticmethod
    def is_na(val: pd.Series) -> pd.Series:
        """
        检查是否为 N/A
        """
        return val.astype(str).eq(CompareConst.N_A)

    @staticmethod
    def rule_num_num(npu_num: pd.Series, bench_num: pd.Series):
        diff = npu_num - bench_num
        rel = pd.Series(CompareConst.INF, index=diff.index)
        mask_nonzero = bench_num != 0
        rel.loc[mask_nonzero] = (diff[mask_nonzero] / bench_num[mask_nonzero] * 100).abs().astype(str) + "%"
        return diff, rel

    @staticmethod
    def static_diff(diff: str, rel=None):
        if rel is None:
            rel = diff
        return diff, rel

    @staticmethod
    def get_number(val):
        return pd.to_numeric(val.astype(str), errors='coerce')

    def build_rules(self):
        """
        创建npu、bench不相等规则
        """
        # NUM × NUM
        self.rules = {(ValType.NUM, ValType.NUM): self.rule_num_num}

        # ---------- NAN 规则 ----------
        nan_rule = self.static_diff(CompareConst.NAN)
        nan_range = {ValType.NUM, ValType.NAN, ValType.INF, ValType.NEG_INF}
        for t in nan_range:
            self.rules[(ValType.NAN, t)] = nan_rule
            self.rules[(t, ValType.NAN)] = nan_rule

        # ---------- INF / -INF 规则 ----------
        pos_inf = self.static_diff(CompareConst.INF)
        neg_inf = self.static_diff(CompareConst.NEG_INF)
        # INF / -INF 在左
        for r in (ValType.NUM, ValType.NEG_INF):
            self.rules[(ValType.INF, r)] = pos_inf
        for r in (ValType.NUM, ValType.INF):
            self.rules[(ValType.NEG_INF, r)] = neg_inf
        # INF / -INF 在右
        self.rules[(ValType.NUM, ValType.INF)] = neg_inf
        self.rules[(ValType.NUM, ValType.NEG_INF)] = pos_inf

        # ---------- N/A 规则 ----------
        na_rule = self.static_diff(CompareConst.N_A)
        na_range = {ValType.NUM, ValType.NAN, ValType.INF, ValType.NEG_INF, ValType.DEVICE, ValType.NA, ValType.OTHER}
        for t in na_range:
            self.rules[(ValType.NA, t)] = na_rule
            self.rules[(t, ValType.NA)] = na_rule

        # ---------- DEVICE / OTHER 规则 ----------
        # 给diff，都是device给N/A
        diff_rule = self.static_diff(CompareConst.DIFF_FLAG)
        diff_range = {ValType.NUM, ValType.NAN, ValType.INF, ValType.NEG_INF, ValType.DEVICE, ValType.OTHER}
        for t in diff_range:
            self.rules[(ValType.DEVICE, t)] = diff_rule
            self.rules[(t, ValType.DEVICE)] = diff_rule
            self.rules[(ValType.OTHER, t)] = diff_rule
            self.rules[(t, ValType.OTHER)] = diff_rule
        self.rules[(ValType.DEVICE, ValType.DEVICE)] = na_rule

    def classify(self, val: pd.Series) -> pd.Series:
        result_values = np.select(
            [
                self.is_nan(val),
                self.is_inf(val),
                self.is_neg_inf(val),
                self.is_number(val),  # 'inf', '-inf'会被pandas认为是number，所以放在inf/-inf判断后面
                self.is_device(val),
                self.is_na(val),
            ],
            [
                ValType.NAN.value,
                ValType.INF.value,
                ValType.NEG_INF.value,
                ValType.NUM.value,
                ValType.DEVICE.value,
                ValType.NA.value,
            ],
            default=ValType.OTHER.value,
        )
        val_to_enum = {t.value: t for t in ValType}
        result_series = pd.Series(result_values, index=val.index).map(val_to_enum)
        return result_series

    def calc_summary_diff(self, result_df, stats_index: str):
        npu_val = result_df['NPU ' + stats_index]
        bench_val = result_df['Bench ' + stats_index]
        diff_name = stats_index.capitalize() + ' diff'
        rel_err_name = ('norm' if stats_index == 'l2norm' else stats_index).capitalize() + 'RelativeErr'

        # ---------------------- 初始化 ----------------------
        result_df[[diff_name, rel_err_name]] = CompareConst.N_A

        # ---------------------- 基础 mask ----------------------
        npu_num = self.get_number(npu_val)
        bench_num = self.get_number(bench_val)
        mask_equal = self.is_same_value(npu_val, bench_val)
        mask_unequal = ~mask_equal

        # ---------------------- npu, bench统计量相等 ----------------------
        result_df.loc[mask_equal, [diff_name, rel_err_name]] = 0

        # ---------------------- npu, bench统计量不相等 ----------------------
        npu_type = self.classify(npu_val)
        bench_type = self.classify(bench_val)
        for t1 in ALL_TYPES:
            for t2 in ALL_TYPES:
                mask = mask_unequal & (npu_type == t1) & (bench_type == t2)
                if not mask.any():
                    continue

                rule = self.rules.get((t1, t2), self.DEFAULT_RULE)

                if callable(rule):
                    diff, rel = rule(npu_num[mask], bench_num[mask])
                else:
                    diff, rel = rule

                result_df.loc[mask, diff_name] = diff
                result_df.loc[mask, rel_err_name] = rel

    def calc_accuracy(self, result_df, header):
        # bench name N/A represents no bench data, err_msg adds "No bench data matched."
        condition_no_bench = result_df[CompareConst.BENCH_NAME] == CompareConst.N_A
        result_df[condition_no_bench] = result_df[condition_no_bench].fillna(CompareConst.N_A)
        result_df.loc[condition_no_bench, CompareConst.ERROR_MESSAGE] = CompareConst.NO_BENCH
        condition_req_grad_consist = result_df[CompareConst.NPU_REQ_GRAD] == result_df[CompareConst.BENCH_REQ_GRAD]

        if self.mode_config.dump_mode == Const.MD5:
            condition_md5_equal = result_df[CompareConst.NPU_MD5] == result_df[CompareConst.BENCH_MD5]
            result_df.loc[condition_md5_equal, CompareConst.RESULT] = CompareConst.PASS
            result_df.loc[~condition_md5_equal & ~condition_no_bench, CompareConst.RESULT] = CompareConst.DIFF
        elif self.mode_config.first_diff_analyze or self.mode_config.dump_mode == Const.SUMMARY:
            for stats_index in ['max', 'min', 'mean', 'l2norm']:
                self.calc_summary_diff(result_df, stats_index)
            result_df.loc[~condition_no_bench, [CompareConst.RESULT, CompareConst.ERROR_MESSAGE]] = ''
            result_df.loc[~condition_req_grad_consist, CompareConst.ERROR_MESSAGE] += 'Requires_grad inconsistent. '
        else:
            fill_cols = [CompareConst.COSINE, CompareConst.EUC_DIST,
                         CompareConst.MAX_ABS_ERR, CompareConst.MAX_RELATIVE_ERR,
                         CompareConst.ONE_THOUSANDTH_ERR_RATIO, CompareConst.FIVE_THOUSANDTHS_ERR_RATIO,
                         CompareConst.ERROR_MESSAGE]
            result_df.loc[~condition_no_bench, fill_cols] = ''  # 默认填充'', df默认省缺值为nan，不便后续处理，容易出现意外情况
            result_df.loc[~condition_req_grad_consist, CompareConst.ERROR_MESSAGE] = 'Requires_grad inconsistent. '

        return result_df[header]


def setup_comparison(input_param, output_path, **kwargs) -> ComparisonConfig:
    """公共的前置处理逻辑，返回封装后的 ComparisonConfig 对象"""
    try:
        config = ComparisonConfig(
            dump_mode='',
            stack_mode=False,
            fuzzy_match=kwargs.get('fuzzy_match', False),
            data_mapping=kwargs.get('data_mapping', {}),
            suffix=kwargs.get('suffix', ''),
            cell_mapping=kwargs.get('cell_mapping', {}),
            api_mapping=kwargs.get('api_mapping', {}),
            layer_mapping=kwargs.get('layer_mapping', {}),
            first_diff_analyze=kwargs.get('first_diff_analyze', False),
            compared_file_type='',
            is_print_compare_log=kwargs.get('is_print_compare_log', False)
        )

        set_dump_path(input_param)
        config.dump_mode = get_dump_mode(input_param)
        config.compared_file_type = get_file_type(input_param.get("npu_path", None))

        # set stack_mode and set "stack_path" in input_param
        config.stack_mode, input_param = set_stack_json_path(input_param)

        check_configuration_param(config)
        create_directory(output_path)
        check_compare_param(input_param, output_path, config.dump_mode, config.stack_mode)

        return config

    except (CompareException, FileCheckException) as error:
        logger.error('Compare failed. Please check the arguments and do it again!')
        raise CompareException(error.code) from error



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

__all__ = ['BaseChecker', 'apply_patches']

import msprobe.core.config_check.checkers.env_args_checker
import msprobe.core.config_check.checkers.pip_checker
import msprobe.core.config_check.checkers.dataset_checker
import msprobe.core.config_check.checkers.weights_checker
import msprobe.core.config_check.checkers.hyperparameter_checker
import msprobe.core.config_check.checkers.random_checker

from msprobe.core.config_check.checkers.base_checker import BaseChecker



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the const value.
"""


class AdvisorConst:
    """
    The class for advisor const
    """
    # column const
    COSINE_SIMILARITY = "CosineSimilarity"
    INDEX = "Index"
    NPU_DUMP = "NPUDump"
    OVERFLOW = "OverFlow"

    # advisor summary key
    DETECTION_TYPE = "Detection Type"
    OPERATOR_INDEX = "Operator Index"
    ADVISOR_SUGGEST = "Expert Advice"

    # detection type
    OVERFLOW_DETECTION = "FP16 Overflow"
    INPUT_DETECTION = "Input Inconsistent"
    CONSISTENCY_DETECTION = "Global Consistency"

    # operator index
    NO_ERROR_OP = "NA"

    # advisor suggest
    OVERFLOW_SUGGEST = "Float16 data overflow occurs. Rectify the fault and perform comparison again."
    INPUT_SUGGEST = "The input data of NPUDump is inconsistent with that of GroundTruth. Use the same data " \
                    "or check the data preprocessing process."
    CONSISTENCY_SUGGEST = "All data in the comparison result meets the accuracy requirements. " \
                          "If data accuracy of the model is still not up to standard in practical application, " \
                          "please check the post-processing process of model outputs."
    PROBLEM_SUGGEST = "The accuracy of some tensors is low, resulting in an unqualified final accuracy. " \
                      "This may be caused by quantization. Calibrate the data or contact Huawei for further diagnosis. "
    DEVIATION_SUGGEST = "The accuracy of some tensors is low, while the final accuracy is qualified. " \
                        "This may be caused by Ascend internal optimization. " \
                        "Ignore or contact Huawei for further diagnosis. "

    # text symbol
    NEW_LINE = "\n"
    COLON = ": "

    ACCURACY_THRESHOLD = 0.99

